from amuse.test.amusetest import TestWithMPI

import tempfile
import numpy

from amuse.community.hermite.interface import Hermite
from amuse.community.kepler.interface import Kepler
from amuse.community.smalln.interface import SmallN

from amuse.units import nbody_system
from amuse.units import units
from amuse.units import constants

from amuse import datamodel
from amuse.ic import plummer
from amuse.couple import multiples
from amuse.couple import encounters
from amuse import io


class TestSimpleMultiples(TestWithMPI):
    previous = None

    def new_smalln(self):
        if not self.previous is None:
            self.previous.stop()

        result = SmallN()
        result.parameters.timestep_parameter = 0.1
        result.parameters.cm_index = 2001
        self.previous = result
        return result

    def new_kepler_si(self):
        unit_converter = nbody_system.nbody_to_si(
            1.0 | units.MSun,
            1.0 | units.AU
        )
        kepler = Kepler(unit_converter)
        kepler.initialize_code()
        return kepler

    def new_kepler(self):
        kepler = Kepler()
        kepler.initialize_code()
        return kepler

    def new_smalln_si(self):

        if not self.previous is None:
            self.previous.stop()
        converter = nbody_system.nbody_to_si(units.MSun, units.parsec)
        result = SmallN(converter)
        result.parameters.timestep_parameter = 0.1
        result.parameters.cm_index = 2001
        return result

    def new_binary(self, mass1, mass2, semi_major_axis,
                   eccentricity=0, keyoffset=-1):
        total_mass = mass1 + mass2
        mass_fraction_particle_1 = mass1 / (total_mass)

        if keyoffset >= 0:
            binary = datamodel.Particles(keys=range(keyoffset, keyoffset+2))
        else:
            binary = datamodel.Particles(2)

        binary[0].mass = mass1
        binary[1].mass = mass2

        mu = nbody_system.G * total_mass

        velocity_perihelion = numpy.sqrt(mu / semi_major_axis * ((1.0 + eccentricity)/(1.0 - eccentricity)))
        radius_perihelion = semi_major_axis * (1.0 - eccentricity)

        binary[0].position = ((1.0 - mass_fraction_particle_1) * radius_perihelion * [1.0, 0.0, 0.0])
        binary[1].position = -(mass_fraction_particle_1 * radius_perihelion * [1.0, 0.0, 0.0])

        binary[0].velocity = ((1.0 - mass_fraction_particle_1) * velocity_perihelion * [0.0, 1.0, 0.0])
        binary[1].velocity = -(mass_fraction_particle_1 * velocity_perihelion * [0.0, 1.0, 0.0])

        return binary

    def create_binaries(self, center_of_mass_particles, mass1, mass2, semi_major_axis,
                   eccentricity=0):
        singles_in_binaries = datamodel.Particles()
        for binary in center_of_mass_particles:
            particles_in_binary = self.new_binary(
                mass1,
                mass2,
                semi_major_axis
            )

            particles_in_binary.radius = semi_major_axis

            binary.child1 = particles_in_binary[0]
            binary.child2 = particles_in_binary[1]
            binary.mass = mass1 + mass2

            particles_in_binary.position += binary.position
            particles_in_binary.velocity += binary.velocity
            singles_in_binaries.add_particles(particles_in_binary)
        return center_of_mass_particles, singles_in_binaries

    def test0(self):
        code = Hermite()
        stars = datamodel.Particles(2)
        stars.mass = 1 | nbody_system.mass
        stars.position = [
            [0.0, 0, 0],
            [1.2, 0, 0]
        ] | nbody_system.length
        stars.velocity = [
            [0.0, 0, 0],
            [0, 0.1, 0]
        ] | nbody_system.speed
        stars.radius = 0.5 | nbody_system.length

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
            interaction_over_code=None
        )
        encounter_code.parameters.hard_binary_factor = 1
        encounter_code.small_scale_factor = 1
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.particles.add_particles(stars)
        multiples_code.commit_particles()
        multiples_code.evolve_model(0.6 | nbody_system.time)
        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.binaries), 1)

    def test1(self):
        code = Hermite()
        stars = datamodel.Particles(keys=(1, 2, 3, 4))
        stars.mass = 1 | nbody_system.mass
        stars.position = [
            [0.0, 0, 0],
            [0.5, 0, 0],
            [2.0, 0, 0],
            [-10.0, 0, 0],
        ] | nbody_system.length
        stars.velocity = [
            [0.0, 0, 0],
            [0, 0.1, 0],
            [0, -0.1, 0],
            [0, 0.2, 0],
        ] | nbody_system.speed
        stars.radius = 0.5 | nbody_system.length

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
            interaction_over_code=None
        )
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.particles.add_particles(stars)
        multiples_code.commit_particles()

        multiples_code.evolve_model(0.6 | nbody_system.time)
        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.binaries), 1)

        self.assertAlmostRelativeEquals(multiples_code.particles[:-1].radius, 0.5 | nbody_system.length)
        self.assertAlmostRelativeEquals(multiples_code.particles[-1].radius, 0.4446 | nbody_system.length, 3)
        multiples_code.evolve_model(2 | nbody_system.time)
        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.binaries), 1)
        multiples_code.evolve_model(3 | nbody_system.time)
        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.particles), 2)
        self.assertEqual(len(multiples_code.binaries), 1)

    def test2(self):
        code = Hermite()
        stars = datamodel.Particles(keys=(1, 2, 3, 4))
        stars.mass = 1 | nbody_system.mass
        stars.position = [
            [0.0, 0, 0],
            [0.5, 0, 0],
            [3, 0, 0],
            [-10, 0, 0],
        ] | nbody_system.length
        stars.velocity = [
            [0.0, 0, 0],
            [0, 0.1, 0],
            [0.0, -0.5, 0],
            [0, 0.2, 0],
        ] | nbody_system.speed
        stars.radius = 0.5 | nbody_system.length

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
            interaction_over_code=None
        )
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.particles.add_particles(stars)
        multiples_code.commit_particles()

        multiples_code.evolve_model(3 | nbody_system.time)
        self.assertEqual(len(multiples_code.multiples), 1)
        print(multiples_code.multiples[0].components)
        self.assertEqual(len(multiples_code.multiples[0].components), 2)
        self.assertEqual(len(multiples_code.particles), 3)
        self.assertEqual(len(multiples_code.binaries), 1)
        self.assertEqual(len(multiples_code.singles), 2)

    def test3(self):
        code = Hermite()
        particles_in_binary = self.new_binary(
            0.1 | nbody_system.mass,
            0.1 | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=1
        )
        particles_in_binary.radius = 0.001 | nbody_system.length
        binary = datamodel.Particle(key=3)
        binary.child1 = particles_in_binary[0]
        binary.child2 = particles_in_binary[1]
        binary.radius = 0.5 | nbody_system.length
        binary.mass = 0.2 | nbody_system.mass
        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
            interaction_over_code=None
        )
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.singles_in_binaries.add_particles(particles_in_binary)
        multiples_code.binaries.add_particle(binary)

        self.assertEqual(len(multiples_code.singles_in_binaries), 2)
        self.assertEqual(id(multiples_code.binaries[0].child1.particles_set), id(multiples_code.singles_in_binaries))

        multiples_code.commit_particles()

        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.components_of_multiples), 2)

    def test4(self):
        code = Hermite()
        stars = datamodel.Particles(keys=(1, 2, 3, 4))
        stars.mass = 1 | nbody_system.mass
        stars.position = [
            [0.0, 0, 0],
            [0.5, 0, 0],
            [2, 0, 0],
            [-10, 0, 0],
        ] | nbody_system.length
        stars.velocity = [
            [0, 0, 0],
            [0, 0.2, 0],
            [0, -0.2, 0],
            [0, 0.3, 0],
        ] | nbody_system.speed
        stars.radius = 0.5 | nbody_system.length

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
            interaction_over_code=None
        )
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.particles.add_particles(stars)
        multiples_code.commit_particles()
        stopping_condition = multiples_code.stopping_conditions.multiples_change_detection
        stopping_condition.enable()

        multiples_code.evolve_model(3 | nbody_system.time)
        self.assertTrue(stopping_condition.is_set())
        self.assertAlmostRelativeEquals(multiples_code.model_time, 0.0075 | nbody_system.time, 4)
        self.assertEqual(len(stopping_condition.particles(0)), 1)
        self.assertEqual(len(stopping_condition.particles(1)), 0)

        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.multiples[0].components), 2)
        self.assertEqual(len(multiples_code.particles), 3)  # 1 multiples with 2 singles, plus 2 singles free
        self.assertEqual(len(multiples_code.binaries), 1)
        self.assertEqual(len(multiples_code.singles), 2)

        multiples_code.evolve_model(3 | nbody_system.time)
        self.assertTrue(stopping_condition.is_set())
        self.assertAlmostRelativeEquals(multiples_code.model_time, 1.2195 | nbody_system.time, 4)
        self.assertEqual(len(stopping_condition.particles(0)), 1)  # 1 new multiple
        self.assertEqual(len(stopping_condition.particles(1)), 1)  # 1 dissolved multiple

        self.assertEqual(len(multiples_code.multiples[0].components), 3)
        self.assertEqual(len(multiples_code.particles), 2)  # 1 multiple, plus 1 single free
        self.assertEqual(len(multiples_code.binaries), 1)
        self.assertEqual(len(multiples_code.singles), 1)

    def test5(self):
        converter = nbody_system.nbody_to_si(units.MSun, units.parsec)

        code = Hermite(converter)
        stars = datamodel.Particles(keys=(1, 2))
        stars.mass = converter.to_si(1 | nbody_system.mass)
        stars.position = converter.to_si([
            [0, 0, 0],
            [1.2, 0, 0]
        ] | nbody_system.length)
        stars.velocity = converter.to_si([
            [0, 0, 0],
            [0, 0.1, 0]
        ] | nbody_system.speed)
        stars.radius = converter.to_si(0.5 | nbody_system.length)

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler_si(),
            resolve_collision_code=self.new_smalln_si(),
            interaction_over_code=None,
            G=constants.G
        )
        encounter_code.parameters.hard_binary_factor = 1
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code,
            G=constants.G
        )
        end_time = converter.to_si(1.0 | nbody_system.time)
        multiples_code.particles.add_particles(stars)
        multiples_code.commit_particles()
        multiples_code.evolve_model(end_time)

        self.assertEqual(len(multiples_code.particles), 1)  # 1 multiples with 2 singles
        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.multiples[0].components), 2)
        self.assertEqual(len(multiples_code.binaries), 1)
        self.assertEqual(len(multiples_code.singles), 0)

    def test6(self):
        converter = nbody_system.nbody_to_si(units.MSun, units.parsec)

        code = Hermite(converter)
        stars = datamodel.Particles(keys=(1, 2, 3, 4))
        stars.mass = converter.to_si(1 | nbody_system.mass)
        stars.position = converter.to_si([
            [0, 0, 0],
            [1.2, 0, 0],
            [100, 0, 0],
            [100, 1.2, 0]
        ] | nbody_system.length)
        stars.velocity = converter.to_si([
            [0, 0, 0],
            [0, 0.1, 0],
            [0, 0, 0],
            [0, 0, 0.1],
        ] | nbody_system.speed)
        stars.radius = converter.to_si(0.5 | nbody_system.length)

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler_si(),
            resolve_collision_code=self.new_smalln_si(),
            interaction_over_code=None,
            G=constants.G
        )
        encounter_code.small_scale_factor = 1.0
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code,
            G=constants.G
        )
        multiples_code.must_handle_one_encounter_per_stopping_condition = False
        multiples_code.particles.add_particles(stars)
        multiples_code.commit_particles()

        stopping_condition = multiples_code.stopping_conditions.multiples_change_detection
        stopping_condition.enable()

        end_time = converter.to_si(3.0 | nbody_system.time)
        print(end_time.as_quantity_in(units.Myr))
        multiples_code.evolve_model(end_time)
        self.assertTrue(stopping_condition.is_set())
        print(multiples_code.model_time.as_quantity_in(units.Myr))
        self.assertAlmostRelativeEquals(multiples_code.model_time, 7.99844 | units.Myr, 4)
        self.assertEqual(len(stopping_condition.particles(0)), 2)
        self.assertEqual(len(stopping_condition.particles(1)), 0)

        self.assertEqual(len(multiples_code.particles), 2)             # 1 multiples with 2 singles
        self.assertEqual(len(multiples_code.multiples), 2)
        self.assertEqual(len(multiples_code.binaries), 2)
        self.assertEqual(len(multiples_code.multiples[0].components), 2)
        self.assertEqual(len(multiples_code.multiples[1].components), 2)
        self.assertEqual(len(multiples_code.singles), 0)
        self.assertEqual(len(multiples_code.all_singles), 4)

    def test7(self):
        converter = nbody_system.nbody_to_si(units.MSun, units.parsec)

        code = Hermite(converter)
        stars = datamodel.Particles(keys=(1, 2))
        stars.mass = converter.to_si(1 | nbody_system.mass)
        stars.position = converter.to_si([
            [0, 0, 0],
            [1.1, 0, 0],
        ] | nbody_system.length)
        stars.velocity = converter.to_si([
            [0, 0, 0],
            [-0.5, 1.5, 0],
        ] | nbody_system.speed)
        stars.radius = converter.to_si(0.55 | nbody_system.length)

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler_si(),
            resolve_collision_code=self.new_smalln_si(),
            interaction_over_code=None,
            G=constants.G
        )
        encounter_code.small_scale_factor = 1.0
        encounter_code.parameters.hard_binary_factor = 1
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code,
            G=constants.G
        )
        multiples_code.must_handle_one_encounter_per_stopping_condition = False
        multiples_code.singles.add_particles(stars)
        multiples_code.commit_particles()

        stopping_condition = multiples_code.stopping_conditions.encounter_detection
        stopping_condition.enable()

        end_time = converter.to_si(3.0 | nbody_system.time)
        print(end_time.as_quantity_in(units.Myr))
        multiples_code.evolve_model(end_time)
        self.assertTrue(stopping_condition.is_set())
        print(multiples_code.model_time.as_quantity_in(units.Myr))
        # self.assertAlmostRelativeEquals(multiples_code.model_time , 5.96955 | units.Myr, 4)
        self.assertEqual(len(stopping_condition.particles(0)), 1)
        model = stopping_condition.particles(0)[0]

        self.assertEqual(len(model.particles_before_encounter), 2)
        self.assertEqual(len(model.particles_after_encounter), 2)

        before = model.particles_before_encounter
        after = model.particles_after_encounter

        self.assertAlmostRelativeEquals(before.center_of_mass(), after.center_of_mass(), 7)
        self.assertAlmostRelativeEquals(before.center_of_mass_velocity(), after.center_of_mass_velocity(), 7)

        total_energy_before = before.kinetic_energy() + before.potential_energy(G=constants.G)
        total_energy_after = after.kinetic_energy() + after.potential_energy(G=constants.G)

        self.assertAlmostRelativeEquals(total_energy_before, total_energy_after, 7)

    def test8(self):
        code = Hermite()
        particles_in_binary = self.new_binary(
            0.1 | nbody_system.mass,
            0.1 | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=1
        )
        particles_in_binary.radius = 0.001 | nbody_system.length
        binary = datamodel.Particle(key=3)
        binary.child1 = particles_in_binary[0]
        binary.child2 = particles_in_binary[1]
        binary.radius = 0.5 | nbody_system.length
        binary.mass = 0.2 | nbody_system.mass
        binary.position = [0.0, 0.0, 0.0] | nbody_system.length
        binary.velocity = [0.0, 0.0, 0.0] | nbody_system.speed
        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
            interaction_over_code=None
        )
        encounter_code.parameters.hard_binary_factor = 1
        encounter_code.small_scale_factor = 1
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.singles_in_binaries.add_particles(particles_in_binary)
        multiples_code.binaries.add_particle(binary)
        multiples_code.must_handle_one_encounter_per_stopping_condition = False

        field_particle = datamodel.Particle(key=4)
        field_particle.mass = 0.5 | nbody_system.mass
        field_particle.radius = 0.1 | nbody_system.length
        field_particle.position = [0.0, 0.2, 0.0] | nbody_system.length
        field_particle.velocity = [0.0, 0.0, 0.0] | nbody_system.speed

        multiples_code.singles.add_particle(field_particle)

        self.assertEqual(len(multiples_code.singles_in_binaries), 2)
        self.assertEqual(id(multiples_code.binaries[0].child1.particles_set), id(multiples_code.singles_in_binaries))

        multiples_code.commit_particles()
        multiples_code.multiples.radius = 0.5 | nbody_system.length
        initial_energy = multiples_code.get_total_energy()

        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.components_of_multiples), 2)
        self.assertEqual(len(multiples_code.particles), 2)

        stopping_condition = multiples_code.stopping_conditions.encounter_detection
        stopping_condition.enable()

        singles = datamodel.Particles()
        singles.add_particles(particles_in_binary)
        singles.add_particle(field_particle)

        singles_energy = singles.kinetic_energy() + singles.potential_energy(G=nbody_system.G)
        self.assertAlmostRelativeEquals(initial_energy, singles_energy, 3)

        multiples_code.evolve_model(2 | nbody_system.time)

        final_energy = multiples_code.get_total_energy()
        self.assertTrue(stopping_condition.is_set())
        self.assertAlmostRelativeEquals(initial_energy, final_energy, 7)

    def test9(self):
        code = Hermite()

        particles_in_binary = self.new_binary(
            0.1 | nbody_system.mass,
            0.1 | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=1
        )
        particles_in_binary.radius = 0.001 | nbody_system.length

        binary = datamodel.Particle(key=3)
        binary.child1 = particles_in_binary[0]
        binary.child2 = particles_in_binary[1]
        binary.radius = 0.5 | nbody_system.length
        binary.mass = 0.2 | nbody_system.mass
        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
        )

        others = datamodel.Particles(key=[4, 5, 6])
        for i in range(3):
            others[i].position = [i, 0, 0] | nbody_system.length
            others[i].velocity = [0, 0, i] | nbody_system.speed
            others[i].mass = 1 | nbody_system.mass
            others[i].radius = 0 | nbody_system.length

        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.singles_in_binaries.add_particles(particles_in_binary)
        multiples_code.binaries.add_particle(binary)

        multiples_code.singles.add_particles(others)

        multiples_code.commit_particles()

        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.components_of_multiples), 2)
        self.assertEqual(len(multiples_code.singles), 3)
        self.assertEqual(len(multiples_code.particles), 4)
        self.assertEqual(len(code.particles), 4)

        self.assertAlmostRelativeEquals(multiples_code.particles[-1].mass, 0.2 | nbody_system.mass)
        self.assertAlmostRelativeEquals(code.particles[-1].mass, 0.2 | nbody_system.mass)
        self.assertAlmostRelativeEquals(code.particles[-1].position, [0, 0, 0] | nbody_system.length, 6)
        self.assertAlmostRelativeEquals(code.particles[-1].velocity, [0, 0, 0] | nbody_system.speed, 6)

        multiples_code.update_model()
        self.assertAlmostRelativeEquals(multiples_code.particles[-1].mass, 0.2 | nbody_system.mass)
        self.assertAlmostRelativeEquals(code.particles[-1].mass, 0.2 | nbody_system.mass)
        self.assertAlmostRelativeEquals(code.particles[-1].position, [0, 0, 0] | nbody_system.length, 6)
        self.assertAlmostRelativeEquals(code.particles[-1].velocity, [0, 0, 0] | nbody_system.speed, 6)

        multiples_code.singles_in_binaries[0].mass = 0.2 | nbody_system.mass

        multiples_code.update_model()

        print(code.particles.mass)
        self.assertAlmostRelativeEquals(multiples_code.particles[-1].mass, 0.3 | nbody_system.mass)
        self.assertAlmostRelativeEquals(code.particles[-1].mass, 0.3 | nbody_system.mass)
        print(code.particles[-1].position)
        print(code.particles[-1].velocity)
        self.assertAlmostRelativeEquals(code.particles[-1].position, [0.00166666666667, 0, 0] | nbody_system.length, 6)
        self.assertAlmostRelativeEquals(code.particles[-1].velocity, [0, 0.7453559925, 0] | nbody_system.speed, 6)

    def test10(self):
        code = Hermite()

        particles_in_binary = self.new_binary(
            0.1 | nbody_system.mass,
            0.1 | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=1
        )
        particles_in_binary.radius = 0.001 | nbody_system.length

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
        )

        encounter_code.parameters.hard_binary_factor = 1
        encounter_code.small_scale_factor = 1

        others = datamodel.Particles(key=[4, 5, 6])
        for i in range(3):
            others[i].position = [i,  0, 0] | nbody_system.length
            others[i].velocity = [0, 0, i] | nbody_system.speed
            others[i].mass = 1 | nbody_system.mass
            others[i].radius = 0.05 | nbody_system.length

        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.must_handle_one_encounter_per_stopping_condition = False
        multiples_code.singles.add_particles(particles_in_binary)
        multiples_code.singles.add_particles(others)

        multiples_code.commit_particles()
        multiples_code.evolve_model(1 | nbody_system.time)

        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.components_of_multiples), 2)
        self.assertEqual(len(multiples_code.singles), 3)
        self.assertEqual(len(multiples_code.particles), 4)
        self.assertEqual(len(code.particles), 4)

        self.assertEqual(id(multiples_code.singles_in_binaries), id(multiples_code.binaries[0].child1.particles_set))
        self.assertEqual(id(multiples_code.components_of_multiples), id(multiples_code.multiples[0].components[0].particles_set))
        # multiples_code.singles_in_binaries[0].mass = 0.2 | nbody_system.mass
        print(multiples_code.particles.mass)
        self.assertAlmostRelativeEquals(multiples_code.particles[-1].mass, 1.1 | nbody_system.mass)
        self.assertAlmostRelativeEquals(multiples_code.particles.mass.sum(), 0.1 + 0.1 + 3.0 | nbody_system.mass)
        multiples_code.update_model()

        self.assertAlmostRelativeEquals(multiples_code.particles[-1].mass, 1.1 | nbody_system.mass)

        index = -1
        if not code.particles[index].mass > 1.0 | nbody_system.mass:
            index = -2
        self.assertAlmostRelativeEquals(code.particles[index].mass, 1.1 | nbody_system.mass)

        multiples_code.singles_in_binaries[0].mass += 0.2 | nbody_system.mass

        multiples_code.update_model()

        self.assertAlmostRelativeEquals(multiples_code.particles[-1].mass, 1.3 | nbody_system.mass)
        self.assertAlmostRelativeEquals(code.particles[index].mass, 1.3 | nbody_system.mass)

    def test11(self):
        code = Hermite()

        particles_in_binary = self.new_binary(
            1.0 | nbody_system.mass,
            1.0 | nbody_system.mass,
            0.001 | nbody_system.length,
            keyoffset=1
        )
        particles_in_binary.radius = 0.01 | nbody_system.length

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
        )

        others = datamodel.Particles(keys=[4, 5, 6])
        for i in range(3):
            others[i].position = [i, 0, 0] | nbody_system.length
            others[i].velocity = [0, 0, 0] | nbody_system.speed
            others[i].mass = 0.2 | nbody_system.mass
            others[i].radius = 0.05 | nbody_system.length

        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.singles.add_particles(particles_in_binary)
        multiples_code.singles.add_particles(others)

        stopping_condition = multiples_code.stopping_conditions.binaries_change_detection
        stopping_condition.enable()

        multiples_code.commit_particles()
        multiples_code.evolve_model(1 | nbody_system.time)
        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.binaries), 1)
        self.assertEqual(len(multiples_code.components_of_multiples), 2)
        self.assertEqual(len(multiples_code.singles), 3)
        self.assertEqual(len(multiples_code.particles), 4)
        self.assertEqual(len(code.particles), 4)
        self.assertTrue(stopping_condition.is_set())
        multiples_code.particles[-1].velocity = [0, 0, 0] | nbody_system.speed
        multiples_code.update_model()
        print(multiples_code.particles.key)

        self.assertEqual(len(stopping_condition.particles(0)), 1)
        self.assertEqual(len(stopping_condition.particles(1)), 0)
        self.assertEqual(len(stopping_condition.particles(2)), 0)
        self.assertAlmostRelativeEquals(multiples_code.multiples[0].mass, 2.0 | nbody_system.mass)
        self.assertAlmostRelativeEquals(multiples_code.particles.mass.sum(), 2.6 | nbody_system.mass)
        print(multiples_code.particles.velocity)
        multiples_code.evolve_model(2 | nbody_system.time)
        self.assertTrue(stopping_condition.is_set())
        self.assertEqual(len(stopping_condition.particles(0)), 0)
        self.assertEqual(len(stopping_condition.particles(1)), 0)
        self.assertEqual(len(stopping_condition.particles(2)), 1)
        self.assertAlmostRelativeEquals(multiples_code.multiples[0].mass, 2.0 | nbody_system.mass)
        self.assertAlmostRelativeEquals(multiples_code.particles.mass.sum(), 2.6 | nbody_system.mass)

    def test12(self):
        code = Hermite()

        particles_in_binary = self.new_binary(
            1.0 | nbody_system.mass,
            1.0 | nbody_system.mass,
            0.001 | nbody_system.length,
            keyoffset=10
        )
        particles_in_binary.radius = 0.01 | nbody_system.length

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
        )
        binary = datamodel.Particle(key=20)
        binary.child1 = particles_in_binary[0]
        binary.child2 = particles_in_binary[1]
        binary.position = [1, 0, 1] | nbody_system.length
        particles_in_binary.position += [1, 0, 1] | nbody_system.length

        others = datamodel.Particles(keys=[4, 5, 6])
        for i in range(3):
            others[i].position = [i*10, 0, 0] | nbody_system.length
            others[i].velocity = [0, 0, 0] | nbody_system.speed
            others[i].mass = 0.2 | nbody_system.mass
            others[i].radius = 0.05 | nbody_system.length

        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.particles.add_particles(others)
        multiples_code.singles_in_binaries.add_particles(particles_in_binary)
        multiples_code.binaries.add_particle(binary)
        multiples_code.commit_particles()
        print(multiples_code.particles)
        self.assertEqual(len(multiples_code.particles), 4)
        self.assertAlmostRelativeEquals(multiples_code.particles[-1].position,  [1, 0, 1] | nbody_system.length)

    def test13(self):
        code = Hermite()

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
        )
        center_of_mass_particles = datamodel.Particles(5)
        center_of_mass_particles.position = (numpy.asarray(range(5))).reshape(5, 1) * ([1.0, 0.0, 0.0] | nbody_system.length)
        center_of_mass_particles.velocity = [0.0, 0.0, 0.0] | nbody_system.speed
        center_of_mass_particles.radius = 0.05 | nbody_system.length
        binaries, singles_in_binaries = self.create_binaries(
            center_of_mass_particles,
            1 | nbody_system.mass,
            0.01 | nbody_system.mass,
            0.0001 | nbody_system.length
        )

        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.singles_in_binaries.add_particles(singles_in_binaries)
        multiples_code.binaries.add_particles(binaries)
        multiples_code.commit_particles()

        # stopping_condition = multiples_code.stopping_conditions.encounter_detection
        # stopping_condition.enable()
        stopping_condition = multiples_code.stopping_conditions.binaries_change_detection
        stopping_condition.enable()
        for x in multiples_code.binaries:
            print(x.key, x.child1.key, x.child2.key)

        multiples_code.evolve_model(1 | nbody_system.time)
        self.assertTrue(stopping_condition.is_set())
        for x in multiples_code.binaries:
            print(x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(0):
            print("NEW:", x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(1):
            print("REMOVED:", x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(2):
            print("UPDATED:", x.key, x.child1.key, x.child2.key)
        for x in multiples_code.singles:
            print(x.key, x.mass)
        self.assertEqual(len(multiples_code.singles_in_binaries) + len(multiples_code.singles), 2*len(center_of_mass_particles))
        self.assertEqual(len(multiples_code.binaries) - len(stopping_condition.particles(0)) + len(stopping_condition.particles(1)),  len(center_of_mass_particles))

    def test14(self):
        code = Hermite()

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
        )
        center_of_mass_particles = datamodel.Particles(5)
        center_of_mass_particles.position = (numpy.asarray(range(5))).reshape(5, 1) * ([1.0, 0.0, 0.0] | nbody_system.length)
        center_of_mass_particles.velocity = [0.0, 0.0, 0.0] | nbody_system.speed
        center_of_mass_particles.radius = 0.05 | nbody_system.length
        binaries, singles_in_binaries = self.create_binaries(
            center_of_mass_particles,
            1 | nbody_system.mass,
            0.1 | nbody_system.mass,
            0.00000001 | nbody_system.length
        )

        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.singles_in_binaries.add_particles(singles_in_binaries)
        multiples_code.binaries.add_particles(binaries)
        multiples_code.commit_particles()

        # stopping_condition = multiples_code.stopping_conditions.encounter_detection
        # stopping_condition.enable()
        stopping_condition = multiples_code.stopping_conditions.binaries_change_detection
        stopping_condition.enable()
        for x in multiples_code.binaries:
            print(x.key, x.child1.key, x.child2.key)

        multiples_code.evolve_model(2 | nbody_system.time)
        self.assertTrue(stopping_condition.is_set())
        for x in multiples_code.binaries:
            print(x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(0):
            print("NEW:", x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(1):
            print("REMOVED:", x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(2):
            print("UPDATED:", x.key, x.child1.key, x.child2.key)
        for x in multiples_code.singles:
            print(x.key, x.mass)
        self.assertEqual(len(multiples_code.singles_in_binaries) + len(multiples_code.singles), 2*len(center_of_mass_particles))
        self.assertEqual(len(multiples_code.binaries) - len(stopping_condition.particles(0)) + len(stopping_condition.particles(1)),  len(center_of_mass_particles))

    def test15(self):
        code = Hermite()

        encounter_code = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=self.new_smalln(),
        )
        n = 10
        center_of_mass_particles = plummer.new_plummer_model(n, random=numpy.random.mtrand.RandomState(1))
        center_of_mass_particles.radius = 0.5 | nbody_system.length
        center_of_mass_particles.velocity *= 0
        binaries, singles_in_binaries = self.create_binaries(
            center_of_mass_particles,
            0.999 * ((1.0 | nbody_system.mass) / n),
            0.001 * ((1.0 | nbody_system.mass) / n),
            0.00001 | nbody_system.length
        )
        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounter_code
        )
        multiples_code.singles_in_binaries.add_particles(singles_in_binaries)
        multiples_code.binaries.add_particles(binaries)
        multiples_code.commit_particles()

        # stopping_condition = multiples_code.stopping_conditions.encounter_detection
        # stopping_condition.enable()
        stopping_condition = multiples_code.stopping_conditions.binaries_change_detection
        stopping_condition.enable()
        for x in multiples_code.binaries:
            print(x.key, x.child1.key, x.child2.key)

        multiples_code.evolve_model(2 | nbody_system.time)
        self.assertTrue(stopping_condition.is_set())
        for x in multiples_code.binaries:
            print(x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(0):
            print("NEW:", x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(1):
            print("REMOVED:", x.key, x.child1.key, x.child2.key)
        for x in stopping_condition.particles(2):
            print("UPDATED:", x.key, x.child1.key, x.child2.key)
        for x in multiples_code.singles:
            print(x.key, x.mass)
        self.assertEqual(len(multiples_code.binaries) - len(stopping_condition.particles(0)) + len(stopping_condition.particles(1)),  len(center_of_mass_particles))

    def test16(self):
        code = Hermite()

        n = 10
        singles = datamodel.Particles(keys=range(1, n+1))
        singles.mass = 1 | nbody_system.mass
        for x in range(n):
            singles[x].position = [x*x, 0, 0] | nbody_system.length
        singles.velocity = [0, 0, 0] | nbody_system.speed

        singles.radius = 0.5 | nbody_system.length

        multiples_code = encounters.Multiples(
            gravity_code=code,
            handle_encounter_code=encounters.StickyHandleEncounter()
        )
        multiples_code.singles.add_particles(singles)
        multiples_code.commit_particles()

        multiples_code.evolve_model(1 | nbody_system.time)
        print(len(multiples_code.multiples))
        self.assertEqual(len(multiples_code.multiples), 1)
        self.assertEqual(len(multiples_code.particles), 9)
        self.assertEqual(len(multiples_code.singles), 8)
        self.assertEqual(len(multiples_code.binaries), 1)
        self.assertEqual(len(multiples_code.singles_in_binaries), 2)
        self.assertEqual(id(multiples_code.components_of_multiples), id(multiples_code.multiples[0].components[0].particles_set))
        print(multiples_code.multiples[0].components)
        with tempfile.NamedTemporaryFile() as temp:
            io.write_set_to_file(
                (
                    multiples_code.singles,
                    multiples_code.singles_in_binaries,
                    multiples_code.binaries,
                    multiples_code.components_of_multiples,
                    multiples_code.multiples
                ),
                temp.name,
                # "multiples.hdf5",
                "hdf5",
                overwrite_file=True,
                version="2.0",
                names=(
                    "singles",
                    "singles_in_binaries",
                    "binaries",
                    "components_of_multiples",
                    "multiples"
                )
                )

            multiples_code_loaded = encounters.Multiples(
                gravity_code=Hermite(),
                handle_encounter_code=encounters.StickyHandleEncounter()
            )

            (
                singles,
                singles_in_binaries,
                binaries,
                components_of_multiples,
                multiples
            ) = io.read_set_from_file(
                temp.name,
                # "multiples.hdf5",
                "hdf5",
                version="2.0",
                names=(
                    "singles",
                    "singles_in_binaries",
                    "binaries",
                    "components_of_multiples",
                    "multiples"
                )
            )
        self.assertEqual(len(multiples), 1)
        self.assertEqual(len(singles), 8)
        self.assertEqual(len(binaries), 1)
        self.assertEqual(len(singles_in_binaries), 2)
        # self.assertEquals(id(components_of_multiples), id(multiples[0].components[0].particles_set))

        multiples_code_loaded.singles.add_particles(singles)
        multiples_code_loaded.singles_in_binaries.add_particles(singles_in_binaries)
        multiples_code_loaded.binaries.add_particles(binaries)
        multiples_code_loaded.components_of_multiples.add_particles(components_of_multiples)
        multiples_code_loaded.multiples.add_particles(multiples)

        multiples_code_loaded.commit_particles()

        self.assertEqual(len(multiples_code_loaded.multiples), 1)
        self.assertEqual(len(multiples_code_loaded.particles), 9)
        self.assertEqual(len(multiples_code_loaded.singles), 8)
        self.assertEqual(len(multiples_code_loaded.binaries), 1)
        self.assertEqual(len(multiples_code_loaded.singles_in_binaries), 2)
        # self.assertEquals(id(multiples_code_loaded.components_of_multiples), id(multiples_code_loaded.multiples[0].components[0].particles_set))

        multiples_code.evolve_model(4 | nbody_system.time)

        # need to use 3 here as the model_time is reset when doing a restart and we dit not set it after creating Hermite
        multiples_code_loaded.evolve_model(3.0 | nbody_system.time)

        print(len(multiples_code.multiples), multiples_code.particles)
        print(multiples_code.particles.position - multiples_code_loaded.particles.position)
        self.assertAlmostRelativeEquals(multiples_code.particles.position - multiples_code_loaded.particles.position, [0, 0, 0] | nbody_system.length)

        for code in [multiples_code, multiples_code_loaded]:
            self.assertEqual(len(code.multiples), 1)
            self.assertEqual(len(code.particles), 8)
            self.assertEqual(len(code.singles), 7)
            self.assertEqual(len(code.binaries), 1)
            self.assertEqual(len(code.singles_in_binaries), 2)
            self.assertEqual(len(code.components_of_multiples), 3)
            self.assertEqual(id(code.components_of_multiples), id(code.multiples[0].components[0].particles_set))



from amuse.test.amusetest import TestWithMPI
import os
import sys
import numpy
import math

from amuse.community.phantom.interface import PhantomInterface, Phantom

from amuse.datamodel import Particles
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.ic import plummer
from amuse.ic.plummer import new_plummer_model

from amuse.test.suite.codes_tests.gd_tests import (
    _TestGravitationalDynamicsInterface,
)
try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


class TestPhantomInterface(TestWithMPI):
    def gravity_code_interface(self):
        return PhantomInterface

    def reference_includes(self):
        return "Price"

    def starting_particle_index(self):
        return 1

    def test_initialise(self):
        interface = self.gravity_code_interface()
        instance = self.new_instance_of_an_optional_code(interface)
        instance.initialize_code()
        instance.stop()

    def test_literature_reference(self):
        interface = self.gravity_code_interface()
        instance = self.new_instance_of_an_optional_code(interface)
        instance.initialize_code()
        reference_string = self.reference_includes()
        self.assertTrue(
            reference_string in instance.all_literature_references_string()
        )
        instance.stop()

    def test_add_and_retrieve_particles(self):
        interface = self.gravity_code_interface()
        instance = self.new_instance_of_an_optional_code(interface)
        instance.initialize_code()
        # Phantom won't work with fewer than 7 particles!
        n = 7
        values = [1.0 * i for i in range(1, n)]
        instance.new_particle(
            values,
            values,
            values,
            values,
            values,
            values,
            values,
        )
        error = instance.commit_particles()
        self.assertEqual(error, 0)
        retrieved_state = instance.get_state(self.starting_particle_index())
        self.assertEqual(1.0, retrieved_state['mass'])
        retrieved_state = instance.get_state(
            self.starting_particle_index()+n-2
        )
        instance.cleanup_code()
        # For any particle other than a sink, Phantom has one fixed mass!
        self.assertEqual(1.0,  retrieved_state['mass'])
        instance.stop()

    def test_add_and_retrieve_sph_particles(self):
        interface = self.gravity_code_interface()
        instance = self.new_instance_of_an_optional_code(interface)
        instance.initialize_code()

        # Phantom won't work with fewer than 7 particles!
        n = 100
        values = [1.0 * i for i in range(1, n)]
        instance.new_sph_particle(
            values,
            values,
            values,
            values,
            values,
            values,
            values,
            values,
        )
        error = instance.commit_particles()
        self.assertEqual(error, 0)
        retrieved_state = instance.get_state(self.starting_particle_index())
        self.assertEqual(1.0, retrieved_state['mass'])
        retrieved_state = instance.get_state(
            self.starting_particle_index()+n-2
        )
        instance.cleanup_code()
        # For any particle other than a sink, Phantom has one fixed mass!
        self.assertEqual(1.0,  retrieved_state['mass'])
        instance.stop()

    def test_parameters(self):
        interface = self.gravity_code_interface()
        # instance = self.new_instance_of_an_optional_code(interface)
        instance = interface(redirection="none")
        instance.initialize_code()
        gamma, error = instance.get_gamma()
        self.assertEqual(0, error)
        self.assertEqual(1., gamma)
        ieos, error = instance.get_ieos()
        self.assertEqual(0, error)
        self.assertEqual(1, ieos)


class TestPhantom(TestWithMPI):
    def test_initialise(self):
        instance = Phantom()
        instance.stop()

    def test_add_gasparticles(self):
        n_particles = 10
        instance = Phantom()
        gas = Particles(n_particles)
        gas.mass = 1 | nbody_system.mass
        gas.x = numpy.arange(n_particles) | nbody_system.length
        gas.y = numpy.arange(n_particles) | nbody_system.length
        gas.z = 0 | nbody_system.length
        gas.velocity = [0, 0, 0] | nbody_system.speed
        gas.u = 0 | nbody_system.speed**2
        instance.gas_particles.add_particles(gas)
        self.assertEqual(10, len(instance.gas_particles))
        instance.stop()



from amuse.datamodel import AbstractSet
from amuse.test.amusetest import TestWithMPI
from amuse.ext import concurrent
from amuse import datamodel

from amuse.rfi.core import *
from amuse.ic.plummer import new_plummer_sphere
from amuse.units import nbody_system
from amuse.units import units

from amuse.community import *
from amuse.support.interface import InCodeComponentImplementation

from amuse.io import read_set_from_file, write_set_to_file

from collections import namedtuple
import numpy
import pickle
try:
    from mpi4py import MPI
except ImportError:
    MPI = None

import base64


def dump_and_encode(x):
    return base64.b64encode(pickle.dumps(x)).decode()


def decode_and_load(x):
    return pickle.loads(base64.b64decode(x.encode()))


class DistributedParticlesInterface(PythonCodeInterface):

    def __init__(self, implementation_factory, **options):
        PythonCodeInterface.__init__(self, implementation_factory=implementation_factory, **options)

    @legacy_function
    def get_length():
        function = LegacyFunctionSpecification()
        function.addParameter('reference', dtype='int64', direction=function.IN)
        function.addParameter('len_out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def do_init():
        function = LegacyFunctionSpecification()
        function.addParameter('size', dtype='int64', direction=function.IN)
        function.addParameter('reference_out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def do_set_attribute():
        function = LegacyFunctionSpecification()
        function.addParameter('reference', dtype='int64', direction=function.IN)
        function.addParameter('name_of_the_attribute', dtype='string', direction=function.IN)
        function.addParameter('pickled_value', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def do_get_attribute():
        function = LegacyFunctionSpecification()
        function.addParameter('reference', dtype='int64', direction=function.IN)
        function.addParameter('name_of_the_attribute', dtype='string', direction=function.IN)
        function.addParameter('output', dtype='string', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def do_getitem():
        function = LegacyFunctionSpecification()
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('pickled_index', dtype='string', direction=function.IN)
        function.addParameter('is_particle_out', dtype='bool', direction=function.OUT)
        function.addParameter('reference_out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_from_generator():
        function = LegacyFunctionSpecification()
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('pickled_generator', dtype='string', direction=function.IN)
        function.addParameter('pickled_args', dtype='string', direction=function.IN)
        function.addParameter('pickled_kwargs', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def select_array():
        function = LegacyFunctionSpecification()
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('pickled_function', dtype='string', direction=function.IN)
        function.addParameter('pickled_attributes', dtype='string', direction=function.IN)
        function.addParameter('reference_out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def write_set_to_file():
        function = LegacyFunctionSpecification()
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('filebase', dtype='string', direction=function.IN)
        function.addParameter('fileformat', dtype='string', direction=function.IN)
        function.addParameter('pickled_filenames', dtype='string', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def read_set_from_file():
        function = LegacyFunctionSpecification()
        function.addParameter('reference_in', dtype='int64', direction=function.IN)
        function.addParameter('pickled_filenames', dtype='string', direction=function.IN)
        function.addParameter('fileformat', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        return function


class DistributedParticlesCode(InCodeComponentImplementation):

    def __init__(self, **options):
        interface = DistributedParticlesInterface(
            implementation_factory=DistributedParticlesImplementation,
            **options
        )

        InCodeComponentImplementation.__init__(
            self,
            interface,
            **options
        )


class DistributedParticles(object):

    def __init__(self, size=0, reference=0, code=None, **options):
        if code is None:
            code = DistributedParticlesCode(**options)
            code.do_init(size)
            reference = 0

        object.__setattr__(self, "code", code)
        object.__setattr__(self, "reference", reference)

    def __len__(self):
        return self.code.get_length(self.reference)

    def __setattr__(self, name_of_the_attribute, value):
        self.code.do_set_attribute(
            self.reference,
            name_of_the_attribute,
            dump_and_encode(value)
        )

    def __getattr__(self, name_of_the_attribute):
        result = self.code.do_get_attribute(self.reference, name_of_the_attribute)
        return decode_and_load(result)

    def __getitem__(self, index):
        is_particle, reference = self.code.do_getitem(
            self.reference,
            dump_and_encode(index)
        )
        if is_particle:
            return DistributedParticle(
                code=self.code,
                reference=reference
            )
        else:
            return DistributedParticles(
                code=self.code,
                reference=reference
            )

    def set_from_generator(self, func, args=(), kwargs={}):
        self.code.set_from_generator(self.reference, dump_and_encode(func),
                dump_and_encode(args), dump_and_encode(kwargs))

    def select_array(self, function, attributes):

        reference = self.code.select_array(self.reference,
                                         dump_and_encode(function),
                                         dump_and_encode(attributes)
        )

        return DistributedParticles(
            code=self.code,
            reference=reference
        )

    def write_set_to_file(self, filebase, fileformat):

        filenames = self.code.write_set_to_file(self.reference, filebase, fileformat)

        return decode_and_load(filenames)

    def read_set_from_file(self, filenames, fileformat):
        if self.reference != 0:
            raise Exception("read only allowed to root set")

        self.code.read_set_from_file(self.reference,
                                     dump_and_encode(filenames),
                                     fileformat)


class DistributedParticle(object):

    def __init__(self, reference, code, **options):
        object.__setattr__(self, "code", code)
        object.__setattr__(self, "reference", reference)

    def __setattr__(self, name_of_the_attribute, value):
        self.code.do_set_attribute(
            self.reference,
            name_of_the_attribute,
            dump_and_encode(value)
        )

    def __getattr__(self, name_of_the_attribute):
        return decode_and_load(self.code.do_get_attribute(self.reference, name_of_the_attribute))


ReferencedParticles = namedtuple('ReferencedParticles', ['particles', 'local_offset', 'local_size', 'global_size'])


class DistributedParticlesImplementation(object):
    EmptyReference = ReferencedParticles(
        None,
        0,
        0,
        0
    )

    def do_init(self, size, reference_out):

        self.mpi_comm = MPI.COMM_WORLD
        self.rank = MPI.COMM_WORLD.Get_rank()
        self.number_of_processes = MPI.COMM_WORLD.Get_size()
        local_size = size // self.number_of_processes
        left_over = size - (local_size * self.number_of_processes)
        if self.rank == 0:
            local_size += left_over
            local_offset = 0
        else:
            local_offset = (local_size * self.rank) + left_over
        self.reference_counter = 0
        self.references_to_particles = {}
        real_particles = datamodel.Particles(local_size)
        self.references_to_particles[self.reference_counter] = ReferencedParticles(
            real_particles,
            local_offset,
            local_size,
            size
        )
        reference_out.value = self.reference_counter
        self.reference_counter += 1
        return 0

    def get_length(self, reference, len_out):
        particles_len = self.references_to_particles[reference].local_size

        input = numpy.zeros(1,  dtype='int64')
        output = numpy.zeros(1,  dtype='int64')

        input[0] = particles_len

        self.mpi_comm.Reduce(
            [input, MPI.INTEGER],
            [output, MPI.INTEGER],
            op=MPI.SUM,
            root=0
        )
        len_out.value = output[0]
        return 0

    def do_set_attribute(self, reference, name_of_the_attribute, pickled_value):
        real_particles, local_offset, local_size, global_size = self.references_to_particles[reference]
        value = decode_and_load(pickled_value)
        if not real_particles is None:
            if (is_quantity(value) and not value.is_scalar()) or hasattr(value, '__iter__'):
                setattr(real_particles, name_of_the_attribute, value[local_offset:local_offset + local_size])
            else:
                setattr(real_particles, name_of_the_attribute, value)
        return 0

    def _merge_results(self, quantities):
        values = []
        value = None
        quantities = [x for x in quantities if x[-1] is not None]

        if len(quantities) == 0:
            return None

        if len(quantities) == 1:
            local_size, local_offset, quantity = quantities[0]
            if local_size == 0:
                return quantity

        total_size = 0
        for local_size, local_offset, quantity in quantities:
            total_size += local_size

        first_quantity = quantities[0][-1]
        result = numpy.zeros(total_size, dtype=first_quantity.dtype)
        if is_quantity(first_quantity):
            result = result | first_quantity.unit

        for local_size, local_offset, quantity in quantities:
            result[local_offset:local_offset+local_size] = quantity
        return result

    def do_get_attribute(self, reference, name_of_the_attribute, output):
        real_particles, local_offset, local_size, global_size = self.references_to_particles[reference]

        try:
            quantity = [local_size, local_offset, getattr(real_particles, name_of_the_attribute)]
        except:
            quantity = [0, 0, None]

        quantities = self.mpi_comm.gather(quantity, root=0)

        if self.rank == 0:
            value = self._merge_results(quantities)
        else:
            value = None
        output.value = dump_and_encode(value)
        return 0

    def do_getitem(self, reference_in, pickled_index, is_particle_out, reference_out):
        real_particles, local_offset, local_size, global_size = self.references_to_particles[reference_in]
        index = decode_and_load(pickled_index)
        self.references_to_particles[self.reference_counter] = self.EmptyReference
        if isinstance(index, int):
            if index >= local_offset and index < local_offset + local_size:
                output_particles = real_particles[index - local_offset]
                self.references_to_particles[self.reference_counter] = ReferencedParticles(
                    output_particles,
                    0,
                    0,
                    0
                )
            is_particle_out.value = True
        elif isinstance(index, slice):
            start, stop, step = index.indices(global_size)
            start -= local_offset
            stop -= local_offset
            start = min(max(0, start), local_size)
            stop = min(max(0, stop), local_size)
            output_particles = real_particles[start:stop]
            input = numpy.zeros(1,  dtype='int64')
            output = numpy.zeros(self.number_of_processes,  dtype='int64')

            input[0] = len(output_particles)

            self.mpi_comm.Allgather(
                [input, MPI.INTEGER],
                [output, MPI.INTEGER]
            )
            total_size = 0
            local_offset = 0
            for i, current_size in enumerate(output):
                if i < self.rank:
                    local_offset += current_size
                total_size += current_size
            referenced_particles = ReferencedParticles(
                output_particles,
                local_offset,
                len(output_particles),
                total_size
            )
            self.references_to_particles[self.reference_counter] = referenced_particles

            is_particle_out.value = False  # check, a slice could still result in 1 particle!!!
        else:
            raise Exception("need to parse index and do smart things here!!")

        reference_out.value = self.reference_counter
        self.reference_counter += 1
        return 0

    def set_from_generator(self, reference, pickled_generator, pickled_args=None, pickled_kwargs=None):
        generator = decode_and_load(pickled_generator)
        args = ()
        if pickled_args is not None:
            args = decode_and_load(pickled_args)
        kwargs = {}
        if pickled_kwargs is not None:
            kwargs = decode_and_load(pickled_kwargs)

        real_particles, local_offset, local_size, global_size = self.references_to_particles[reference]
        particles = generator(local_offset, local_offset+local_size, *args, **kwargs)
        keys = real_particles.get_all_keys_in_store()
        real_particles.remove_particles(real_particles[:])
        attributes = particles.get_attribute_names_defined_in_store()
        indices = particles.get_all_indices_in_store()
        values = particles.get_values_in_store(indices, attributes)
        result = datamodel.Particles()
        result.add_particles_to_store(keys, attributes, values)
        real_particles.add_particles(result)
        return 0

    def select_array(self, reference_in, pickled_function, pickled_attributes, reference_out):
        real_particles, local_offset, local_size, global_size = self.references_to_particles[reference_in]
        function = decode_and_load(pickled_function)
        attributes = decode_and_load(pickled_attributes)

        output_particles = real_particles.select_array(function, attributes)

        input = numpy.zeros(1,  dtype='int64')
        output = numpy.zeros(self.number_of_processes,  dtype='int64')

        input[0] = len(output_particles)

        self.mpi_comm.Allgather(
                [input, MPI.INTEGER],
                [output, MPI.INTEGER]
        )
        total_size = 0
        local_offset = 0
        for i, current_size in enumerate(output):
            if i < self.rank:
                local_offset += current_size
            total_size += current_size
        referenced_particles = ReferencedParticles(
            output_particles,
            local_offset,
            len(output_particles),
            total_size
        )
        self.references_to_particles[self.reference_counter] = referenced_particles
        reference_out.value = self.reference_counter
        self.reference_counter += 1
        return 0

    def write_set_to_file(self, reference_in, filebase, fileformat, filenames_out):
        real_particles, local_offset, local_size, global_size = self.references_to_particles[reference_in]
        filename = filebase+"_%6.6i" % self.rank
        write_set_to_file(real_particles, filename, fileformat,
           local_offset=local_offset, local_size=local_size, global_size=global_size, overwrite_file=True)
        filenames = self.mpi_comm.gather(filename, root=0)
        if self.rank == 0:
            value = filenames
        else:
            value = None
        filenames_out.value = dump_and_encode(value)
        return 0

    def read_set_from_file(self, reference_in, filenames_in, fileformat):
        ceil = lambda x, y: (x//y+(x % y > 0))

        filenames = decode_and_load(filenames_in)
        if reference_in != 0:
            raise Exception("reference_in hould be zero here!")
        real_particles, local_offset, local_size, global_size = self.references_to_particles[reference_in]
        start = self.rank*ceil(len(filenames), self.number_of_processes)
        end = (self.rank+1)*ceil(len(filenames), self.number_of_processes)
        for i in range(start, end):
            if i < len(filenames):
                p = read_set_from_file(filenames[i], fileformat)
                real_particles.add_particles(p)

        input = numpy.zeros(1,  dtype='int64')
        output = numpy.zeros(self.number_of_processes,  dtype='int64')

        input[0] = len(real_particles)
        self.mpi_comm.Allgather(
                [input, MPI.INTEGER],
                [output, MPI.INTEGER]
        )
        total_size = 0
        local_offset = 0
        for i, current_size in enumerate(output):
            if i < self.rank:
                local_offset += current_size
            total_size += current_size
        referenced_particles = ReferencedParticles(
            real_particles,
            local_offset,
            len(real_particles),
            total_size
        )
        self.references_to_particles[reference_in] = referenced_particles
        return 0


def generate_set_example_function(start, end, *args, **kwargs):
    from amuse.datamodel import Particles
    p = Particles(end-start)
    p.index = range(start, end)
    return p


def distributed_king_generator(start, end, *args, **kwargs):
    from amuse.ic.kingmodel import MakeKingModel
    import numpy
    numpy.random.seed(1234567)
    numpy.random.uniform(size=start*6)
    total_number_of_particles = args[0]

    class PartialKingModel(MakeKingModel):
        def makeking(self):
            m, p, v = MakeKingModel.makeking(self)
            m = 0.*m+(1.0 / total_number_of_particles)
            return m, p, v
    kwargs['center_model'] = False
    kwargs['do_scale'] = False
    args = (end-start,)+args[1:]
    return PartialKingModel(*args, **kwargs).result


def select_example_function(x):
    return x > 5


class TestDistributedParticles(TestWithMPI):

    def setUp(self):
        if MPI is None or CodeInterface(must_start_worker=False).channel_type != 'mpi':
            self.skip("test needs mpi")

    def test1(self):
        x = DistributedParticles(
            size=3,
            number_of_workers=1
        )
        self.assertEqual(len(x), 3)
        x.mass = 10 | units.MSun
        self.assertEqual(x.mass, [10, 10, 10] | units.MSun)

    def test2(self):
        x = DistributedParticles(
            size=3,
            number_of_workers=1
        )
        self.assertEqual(len(x), 3)
        x.mass = 10 | units.MSun
        y = x[0:2]
        self.assertEqual(len(y), 2)
        self.assertEqual(y.mass, [10, 10] | units.MSun)

    def test3(self):
        x = DistributedParticles(
            size=3,
            number_of_workers=1
        )
        self.assertEqual(len(x), 3)
        x.mass = [1, 2, 3] | units.MSun
        y = x[0:2]
        z = x[1:]
        self.assertEqual(len(y), 2)
        self.assertEqual(y.mass, [1, 2] | units.MSun)
        self.assertEqual(len(z), 2)
        self.assertEqual(z.mass, [2, 3] | units.MSun)
        z.mass = [4, 5] | units.MSun
        self.assertEqual(y.mass, [1, 4] | units.MSun)
        self.assertEqual(z.mass, [4, 5] | units.MSun)
        self.assertEqual(x.mass, [1, 4, 5] | units.MSun)

    def test4(self):
        x = DistributedParticles(
            size=3,
            number_of_workers=1
        )
        x.mass = [1, 2, 3] | units.MSun
        y = x[1]
        self.assertEqual(y.mass, 2 | units.MSun)
        y.mass = 10 | units.MSun
        self.assertEqual(y.mass, 10 | units.MSun)
        self.assertEqual(x.mass, [1, 10, 3] | units.MSun)

    def test5(self):
        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        self.assertEqual(len(x), 8)
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        for index in range(len(x)):
            self.assertEqual(x[index].mass, (index+1) | units.MSun)

    def test6(self):
        x = DistributedParticles(
            size=9,
            number_of_workers=2
        )
        self.assertEqual(len(x), 9)
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8, 9] | units.MSun
        for index in range(len(x)):
            self.assertEqual(x[index].mass, (index+1) | units.MSun)

    def test7(self):
        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        self.assertEqual(len(x), 8)
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        self.assertEqual(len(x[3:7]), 4)
        x[3:7].mass = [10, 11, 12, 13] | units.MSun
        expected = [1, 2, 3, 10, 11, 12, 13, 8] | units.MSun
        for index in range(len(x)):
            self.assertEqual(x[index].mass, expected[index])

    def test8(self):
        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        self.assertEqual(len(x), 8)
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        self.assertEqual(len(x[3:7]), 4)
        x[3:7].mass = [10, 11, 12, 13] | units.MSun

        self.assertEqual(x[2:6].mass, [3, 10, 11, 12] | units.MSun)

    def test9(self):
        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        self.assertEqual(len(x), 8)
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8]
        self.assertEqual(len(x[3:7]), 4)
        x[3:7].mass = [10, 11, 12, 13]

        self.assertEqual(x[2:6].mass, [3, 10, 11, 12])

    def test10(self):
        x = DistributedParticles(
            size=40,
            number_of_workers=4
        )
        self.assertEqual(len(x), 40)
        x.mass = range(40)
        self.assertEqual(len(x[15:25]), 10)
        self.assertEqual(x[15:25].mass, list(range(15, 25)))
        x[15:25].mass = range(10)
        self.assertEqual(x[15:25].mass, list(range(10)))

    def test11(self):
        from .test_distributed_particles import generate_set_example_function
        y = generate_set_example_function(0, 10)

        x = DistributedParticles(
            size=10,
            number_of_workers=2
        )
        x.set_from_generator(generate_set_example_function)
        self.assertEqual(y.index, x.index)

    def test12(self):
        from .test_distributed_particles import generate_set_example_function
        from .test_distributed_particles import select_example_function
        y = generate_set_example_function(0, 10)

        x = DistributedParticles(
            size=10,
            number_of_workers=2
        )
        x.set_from_generator(generate_set_example_function)
        self.assertEqual(y.index, x.index)

        highy = y.select_array(select_example_function, ("index",))
        highx = x.select_array(select_example_function, ("index",))
        self.assertEqual(highy.index, highx.index)

    def test13(self):
        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        x.index = [0, 10, 10, 10, 0, 0, 10, 0]
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        highx = x.select_array(select_example_function, ("index",))
        self.assertEqual(highx.mass, [2, 3, 4, 7] | units.MSun)

    def test14(self):
        test_results_path = self.get_path_to_results()
        output_files = []
        filebase = os.path.join(test_results_path, "test_distributed_sets")
        for i in [0, 1]:
            output_file = filebase+"_%6.6i" % i
            output_files.append(output_file)

            if os.path.exists(output_file):
                os.remove(output_file)

        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        x.index = [0, 10, 10, 10, 0, 0, 10, 0]
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        files = x.write_set_to_file(filebase, "amuse")
        self.assertEqual(files, output_files)
        for i, f in enumerate(files):
            self.assertTrue(os.path.isfile(f))
            p = read_set_from_file(f, "amuse")
            self.assertEqual(len(p), 4)
            self.assertEqual(p.collection_attributes.global_size, 8)
            self.assertEqual(p.collection_attributes.local_size, 4)

    def test15(self):
        test_results_path = self.get_path_to_results()
        output_files = []
        filebase = os.path.join(test_results_path, "test_distributed_sets")
        for i in [0, 1]:
            output_file = filebase+"_%6.6i" % i
            output_files.append(output_file)

            if os.path.exists(output_file):
                os.remove(output_file)

        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        x.index = [0, 10, 10, 10, 0, 0, 10, 0]
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        highx = x.select_array(select_example_function, ("index",))
        files = highx.write_set_to_file(filebase, "amuse")
        self.assertEqual(files, output_files)
        expected_local_sizes = [3, 1]
        for i, f in enumerate(files):
            self.assertTrue(os.path.isfile(f))
            p = read_set_from_file(f, "amuse")
            self.assertEqual(len(p), expected_local_sizes[i])
            self.assertEqual(p.collection_attributes.global_size, 4)
            self.assertEqual(p.collection_attributes.local_size, expected_local_sizes[i])

    def test16(self):
        test_results_path = self.get_path_to_results()
        filebase = os.path.join(test_results_path, "test_distributed_sets")
        for i in [0, 1]:
            output_file = filebase+"_%6.6i" % i

            if os.path.exists(output_file):
                os.remove(output_file)

        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        x.index = [0, 10, 10, 10, 0, 0, 10, 0]
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        files = x.write_set_to_file(filebase, "amuse")

        y = DistributedParticles(
            size=0,
            number_of_workers=2
        )

        y.read_set_from_file(files, "amuse")
        self.assertEqual(len(y), len(x))
        self.assertEqual(y.index, x.index)
        self.assertEqual(y.mass, x.mass)

    def test17(self):
        test_results_path = self.get_path_to_results()
        filebase = os.path.join(test_results_path, "test_distributed_sets")
        for i in [0, 1]:
            output_file = filebase+"_%6.6i" % i

            if os.path.exists(output_file):
                os.remove(output_file)

        x = DistributedParticles(
            size=8,
            number_of_workers=4
        )
        x.index = [0, 10, 10, 10, 0, 0, 10, 0]
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        files = x.write_set_to_file(filebase, "amuse")

        y = DistributedParticles(
            size=0,
            number_of_workers=2
        )

        y.read_set_from_file(files, "amuse")
        self.assertEqual(len(y), len(x))
        self.assertEqual(y.index, x.index)
        self.assertEqual(y.mass, x.mass)

# number of workers > number of files
# still problematic, because of non-existing attributes if nothing read
    def test18(self):
        test_results_path = self.get_path_to_results()
        filebase = os.path.join(test_results_path, "test_distributed_sets")
        for i in [0, 1]:
            output_file = filebase+"_%6.6i" % i

            if os.path.exists(output_file):
                os.remove(output_file)

        x = DistributedParticles(
            size=8,
            number_of_workers=2
        )
        x.index = [0, 10, 10, 10, 0, 0, 10, 0]
        x.mass = [1, 2, 3, 4, 5, 6, 7, 8] | units.MSun
        files = x.write_set_to_file(filebase, "amuse")

        z = DistributedParticles(
            size=0,
            number_of_workers=4
        )

        z.read_set_from_file(files, "amuse")
        self.assertEqual(len(x), len(z))
        self.assertEqual(x.index, z.index)
        self.assertEqual(x.mass, z.mass)

    def test19(self):
        from .test_distributed_particles import distributed_king_generator
        from amuse.ic.kingmodel import MakeKingModel

        N = 100
        W0 = 7.
        numpy.random.seed(1234567)
        y = MakeKingModel(N, W0, center_model=False).result

        x = DistributedParticles(
            size=N,
            number_of_workers=4
        )
        x.set_from_generator(distributed_king_generator, args=(N, W0))
        self.assertEqual(y.mass, x.mass)
        self.assertEqual(y.x, x.x)
        self.assertEqual(y.y, x.y)
        self.assertEqual(y.z, x.z)
        self.assertEqual(y.vx, x.vx)
        self.assertEqual(y.vy, x.vy)
        self.assertEqual(y.vz, x.vz)



import random
import numpy.random
import sys

from amuse.test import amusetest
from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system
from amuse.support.interface import ConvertArgumentsException

from amuse.ic.plummer import new_plummer_sphere
from amuse.ic.salpeter import new_salpeter_mass_distribution_nbody
from amuse.datamodel import Particle, Particles, ParticlesWithUnitsConverted
from amuse.datamodel import particle_attributes


class TestParticlesAttributes(amusetest.TestCase):

    def test1(self):
        print("Test basic particle attributes and scale_to_standard - nbody units")
        particles = Particles(2)
        particles.position = [[-1, 0, 0], [1, 0, 0]] | nbody_system.length
        particles.velocity = [[-1, 0, 0], [1, 0, 0]] | nbody_system.length/nbody_system.time
        particles.mass = 0.4 | nbody_system.mass
        self.assertAlmostRelativeEquals(particles.total_mass(), 0.8 | nbody_system.mass)
        self.assertAlmostRelativeEquals(particles.kinetic_energy(), 0.4 | nbody_system.energy)
        self.assertAlmostRelativeEquals(particles.potential_energy(G=nbody_system.G), -0.08 | nbody_system.energy)
        self.assertAlmostRelativeEquals(particles.virial_radius(), 4.0 | nbody_system.length)
        particles.scale_to_standard()
        self.assertAlmostRelativeEquals(particles.total_mass(), 1.0 | nbody_system.mass)
        self.assertAlmostRelativeEquals(particles.kinetic_energy(), 0.25 | nbody_system.energy)
        self.assertAlmostRelativeEquals(particles.potential_energy(G=nbody_system.G), -0.5 | nbody_system.energy)
        self.assertAlmostRelativeEquals(particles.virial_radius(), 1.0 | nbody_system.length)

        particles.scale_to_standard(virial_ratio=1)  # unbound
        self.assertAlmostRelativeEquals(particles.kinetic_energy(), 0.5 | nbody_system.energy)
        self.assertAlmostRelativeEquals(particles.potential_energy(G=nbody_system.G), -0.5 | nbody_system.energy)
        particles.scale_to_standard(virial_ratio=0)  # velocities zeroed
        self.assertAlmostRelativeEquals(particles.kinetic_energy(), 0 | nbody_system.energy)
        self.assertAlmostRelativeEquals(particles.potential_energy(G=nbody_system.G), -0.5 | nbody_system.energy)

    def test2(self):
        print("Test basic particle attributes and scale_to_standard - SI units")
        convert_nbody = nbody_system.nbody_to_si(1 | units.MSun, 1 | units.parsec)
        particles = Particles(2)
        particles.position = [[-1, 0, 0], [1, 0, 0]] | units.parsec
        particles.velocity = [[-1, 0, 0], [1, 0, 0]] | units.parsec / units.Myr
        particles.mass = 0.5 | units.MSun

        self.assertAlmostRelativeEquals(particles.total_mass(), 1.0 | units.MSun)
        self.assertAlmostRelativeEquals(particles.kinetic_energy(), 1.0 * (0.5 | units.MSun) * (1 | units.parsec / units.Myr) ** 2)
        self.assertAlmostRelativeEquals(particles.potential_energy(), -constants.G * (0.5 | units.MSun) ** 2 / ([2, 0, 0] | units.parsec).length())
        self.assertAlmostRelativeEquals(particles.virial_radius(), 4.0 | units.parsec)

        particles.scale_to_standard(convert_nbody)
        self.assertAlmostRelativeEquals(particles.total_mass(), convert_nbody.to_si(1.0 | nbody_system.mass))
        self.assertAlmostRelativeEquals(particles.kinetic_energy(), convert_nbody.to_si(0.25 | nbody_system.energy))
        self.assertAlmostRelativeEquals(particles.potential_energy().as_quantity_in(units.J), convert_nbody.to_si(-0.5 | nbody_system.energy).as_quantity_in(units.J), 12)
        self.assertAlmostRelativeEquals(particles.virial_radius(), convert_nbody.to_si(1.0 | nbody_system.length))

        particles.scale_to_standard(convert_nbody, virial_ratio=1)  # unbound
        self.assertAlmostRelativeEquals(particles.kinetic_energy(), 0.5 * constants.G * (1 | units.MSun**2 / units.parsec), 13)
        self.assertAlmostRelativeEquals(particles.potential_energy(), -0.5 * constants.G * (1 | units.MSun**2 / units.parsec))
        particles.scale_to_standard(convert_nbody, virial_ratio=0)  # velocities zeroed
        self.assertAlmostRelativeEquals(particles.kinetic_energy(), 0 | units.J)
        self.assertAlmostRelativeEquals(particles.potential_energy(), -0.5 * constants.G * (1 | units.MSun**2 / units.parsec))

    def test7(self):
        print("Test minimum_spanning_tree_length")
        particles = Particles(6)
        particles.position = [[-5, 0, 0], [-1, 0, 0], [0, 0, 0], [0, 1, 0], [0, -2, 0], [-1, 0.1, 0]] | units.m
        self.assertEqual(particles[:1].minimum_spanning_tree_length(), 0 | units.m)
        self.assertEqual(particles[:2].minimum_spanning_tree_length(), 4 | units.m)
        self.assertEqual(particles[:3].minimum_spanning_tree_length(), 5 | units.m)
        self.assertEqual(particles[:4].minimum_spanning_tree_length(), 6 | units.m)
        self.assertEqual(particles[:5].minimum_spanning_tree_length(), 8 | units.m)
        self.assertEqual(particles[:6].minimum_spanning_tree_length(), 8.1 | units.m)

    def test8(self):
        print("Test mass_segregation_ratio")
        numpy.random.seed(123)
        random.seed(456)
        number_of_particles = 10000
        particles = new_plummer_sphere(number_of_particles)
        particles.r_squared = particles.position.lengths_squared()
        sorted = particles.sorted_by_attribute("r_squared")

        sorted.mass = numpy.random.uniform(1.0, 2.0, number_of_particles) | nbody_system.mass
        MSR = sorted.mass_segregation_ratio(number_of_particles=10, number_of_random_sets=10)

        if sys.hexversion > 0x03000000:
            self.assertAlmostEqual(MSR, 0.7160, 3)
        else:
            self.assertAlmostEqual(MSR, 0.8877, 3)

        random.seed(456)
        result = sorted.mass_segregation_ratio(number_of_particles=10, number_of_random_sets=10, also_compute_uncertainty=True)
        self.assertTrue(isinstance(result, particle_attributes.MassSegregationRatioResults))
        if sys.hexversion > 0x03000000:
            self.assertAlmostEqual(result.mass_segregation_ratio, 0.7160, 3)
            self.assertAlmostEqual(result.uncertainty, 0.2321, 3)
        else:
            self.assertAlmostEqual(result.mass_segregation_ratio, 0.8877, 3)
            self.assertAlmostEqual(result.uncertainty, 0.2482, 3)
        MSR, sigma = sorted.mass_segregation_ratio(number_of_particles=10, number_of_random_sets=50, also_compute_uncertainty=True)
        self.assertTrue(MSR - sigma < 1.0 < MSR + sigma)

        sorted.mass = numpy.linspace(1.0, 2.0, number_of_particles) | nbody_system.mass
        MSR, sigma = sorted.mass_segregation_ratio(number_of_particles=10, number_of_random_sets=20,
            also_compute_uncertainty=True)
        self.assertTrue(MSR < 0.1)
        self.assertTrue(sigma < MSR)

        sorted.mass = numpy.linspace(2.0, 1.0, number_of_particles) | nbody_system.mass
        MSR, sigma = sorted.mass_segregation_ratio(number_of_particles=10, number_of_random_sets=20,
            also_compute_uncertainty=True)
        self.assertTrue(MSR > 10.0)
        self.assertTrue(sigma < MSR)

    def test9(self):
        print("Test __doc__ and help for particle attributes")
        particles = Particles(2)
        self.assertTrue("Returns the total kinetic energy of the\n    particles in the particles set." in particles.kinetic_energy.__doc__)
        self.assertEqual(particles.kinetic_energy.__class__.__name__, "BoundParticlesFunctionAttribute")
        self.assertEqual(particles.kinetic_energy.__class__._function.__name__, "kinetic_energy")
        # __doc__ must be defined on the __class__ for the Python help to work:
        self.assertTrue("Returns the total kinetic energy of the\n    particles in the particles set." in particles.kinetic_energy.__class__.__doc__)

        # difference between particle function and particleS function:
        self.assertTrue("Returns the specific kinetic energy of each particle in the set." in particles.specific_kinetic_energy.__doc__)
        self.assertTrue("Returns the specific kinetic energy of each particle in the set." in particles.specific_kinetic_energy.__class__.__doc__)
        self.assertTrue("Returns the specific kinetic energy of the particle." in particles[0].specific_kinetic_energy.__doc__)
        self.assertTrue("Returns the specific kinetic energy of the particle." in particles[0].specific_kinetic_energy.__class__.__doc__)

        self.assertTrue("Returns the potential at the position of each particle in the set." in particles.potential.__doc__)
        self.assertTrue("Returns the potential at the position of each particle in the set." in particles.potential.__class__.__doc__)
        self.assertTrue("Returns the potential at the position of the particle." in particles[0].potential.__doc__)
        self.assertTrue("Returns the potential at the position of the particle." in particles[0].potential.__class__.__doc__)

    def test10(self):
        particles = Particles(2)
        particles.position = [[1, 0, 0], [2, 0, 0]] | units.m
        particles.velocity = [[3, 0, 0], [4, 0, 0]] | units.m / units.s
        particles.mass = 1 | units.kg

        self.assertEqual(particles.total_mass(), 2 | units.kg)
        self.assertEqual(particles.total_momentum(), [7, 0, 0] | units.kg * units.m / units.s)
        self.assertEqual(particles.total_momentum(), particles.total_mass() * particles.center_of_mass_velocity())
        self.assertEqual(particles.total_radius(), 0.5 | units.m)

        convert_nbody = nbody_system.nbody_to_si(1000 | units.kg, 1e-6 | units.m)
        numpy.random.seed(123)
        field = new_plummer_sphere(10000, convert_nbody)  # small clump of particles, can be regarded as point mass
        self.assertAlmostRelativeEquals(particles.potential_energy_in_field(field), -constants.G * (1500 | units.kg**2 / units.m), 5)
        self.assertAlmostEqual(particles.potential_energy_in_field(field), -1.001142 | 1e-7 * units.kg * units.m**2 / units.s**2, 5)

        field.position *= ((5 | units.m) / field.position.lengths()).reshape((-1, 1))  # spherical shell around particles
        potential_energy = particles.potential_energy_in_field(field)
        particles.position += [0, 1, 2] | units.m  # as long as particles remain inside the shell, the potential doesn't change
        self.assertAlmostEqual(particles.potential_energy_in_field(field), potential_energy, 5)

        particles.mass = [1, 2] | units.kg
        self.assertAlmostRelativeEquals(particles.potential(), -constants.G * ([2, 1] | units.kg / units.m))
        self.assertAlmostRelativeEquals(particles.potential()[0], particles[0].potential())
        self.assertAlmostRelativeEquals(particles.potential()[1], particles[1].potential())

    def test11(self):
        print("Test nearest_neighbour")
        particles = Particles(21)
        particles.x = numpy.logspace(0.0, 2.0, 21) | units.m
        particles.y = 0.0 | units.m
        particles.z = 0.0 | units.m
        self.assertEqual(particles.nearest_neighbour()[0], particles[1])
        self.assertEqual(particles.nearest_neighbour()[1:].key, particles[:-1].key)

        neighbours = Particles(3)
        neighbours.x = [1.0, 10.0, 100.0] | units.m
        neighbours.y = 0.0 | units.m
        neighbours.z = 0.0 | units.m
        self.assertEqual(particles.nearest_neighbour(neighbours).key, neighbours.key[[0]*8 + [1]*10 + [2]*3])

        # A few tests to check the correct behaviour of 'max_array_length' (to prevent memory overflow)
        nearest_neighbours = particles.nearest_neighbour(max_array_length=3*21*21)  # all in one go
        self.assertEqual(nearest_neighbours[0], particles[1])
        self.assertEqual(nearest_neighbours[1:].key, particles[:-1].key)
        nearest_neighbours = particles.nearest_neighbour(max_array_length=3*21*21-1)  # two passes
        self.assertEqual(nearest_neighbours[0], particles[1])
        self.assertEqual(nearest_neighbours[1:].key, particles[:-1].key)
        nearest_neighbours = particles.nearest_neighbour(max_array_length=1)  # 21 passes, one for each particle
        self.assertEqual(nearest_neighbours[0], particles[1])
        self.assertEqual(nearest_neighbours[1:].key, particles[:-1].key)
        self.assertEqual(particles.nearest_neighbour(neighbours, max_array_length=189).key, neighbours.key[[0]*8 + [1]*10 + [2]*3])  # all in one go
        self.assertEqual(particles.nearest_neighbour(neighbours, max_array_length=188).key, neighbours.key[[0]*8 + [1]*10 + [2]*3])  # two passes
        self.assertEqual(particles.nearest_neighbour(neighbours, max_array_length=1).key, neighbours.key[[0]*8 + [1]*10 + [2]*3])  # 21 passes, one for each particle

    def new_koch_star(self, level=5):
        height = numpy.sqrt(3) / 6.0

        def next_iteration(x_values, y_values):
            dx = x_values[1:] - x_values[:-1]
            dy = y_values[1:] - y_values[:-1]
            x_one_third = x_values[:-1] + dx / 3.0
            y_one_third = y_values[:-1] + dy / 3.0
            x_two_third = x_one_third + dx / 3.0
            y_two_third = y_one_third + dy / 3.0
            x_new_point = x_values[:-1] + 0.5*dx - height * dy
            y_new_point = y_values[:-1] + 0.5*dy + height * dx
            new_x = numpy.append(numpy.dstack((x_values[:-1], x_one_third, x_new_point, x_two_third)), [x_values[-1]])
            new_y = numpy.append(numpy.dstack((y_values[:-1], y_one_third, y_new_point, y_two_third)), [y_values[-1]])
            return new_x, new_y
        x, y = numpy.array([0.0, 0.5, 1.0, 0.0]), numpy.array([0.0, 3*height, 0.0, 0.0])
        for i in range(level):
            x, y = next_iteration(x, y)
        return x, y

    def test12(self):
        print("Test correlation_dimension")
        # Particles distributed uniformly in 3D
        particles = Particles(729)
        particles.position = numpy.mgrid[0:9.0, 0:9.0, 0:9.0].reshape(3, -1).transpose() | units.m
        dimension = particles.correlation_dimension()
        self.assertAlmostRelativeEquals(dimension, 3.0, 1)
        # Fractal dimension is scale-free
        particles.position *= 1000
        self.assertAlmostRelativeEquals(dimension, particles.correlation_dimension(), 10)

        # Particles distributed in the x-y plane
        particles.position = numpy.concatenate((numpy.mgrid[0:27.0, 0:27.0].reshape(2, -1), numpy.zeros((1, 729)))).transpose() | units.m
        dimension = particles.correlation_dimension()
        self.assertAlmostRelativeEquals(dimension, 2.0, 1)
        particles.position *= 1000
        self.assertAlmostRelativeEquals(dimension, particles.correlation_dimension(), 10)

        # Particles distributed along a line
        particles.position = numpy.concatenate([numpy.arange(729.0).reshape(1, -1)]*3).transpose() | units.m
        dimension = particles.correlation_dimension()
        self.assertAlmostRelativeEquals(dimension, 1.0, 1)
        particles.position *= 1000
        self.assertAlmostRelativeEquals(dimension, particles.correlation_dimension(), 10)

        # Particles on a Koch curve
        x, y = self.new_koch_star(level=6)
        numpy.random.seed(123456)
        sel = numpy.random.randint(len(x), size=729)
        particles.position = numpy.hstack((x[sel], y[sel], numpy.zeros(729))) | units.m
        dimension = particles.correlation_dimension()
        self.assertAlmostRelativeEquals(dimension, 1.26186, 1)
        particles.position *= 1000
        self.assertAlmostRelativeEquals(dimension, particles.correlation_dimension(), 10)

    def test13(self):
        print("Test box_counting_dimension")
        # Particles distributed uniformly in 3D
        particles = Particles(4096)
        particles.position = numpy.mgrid[0:16.0, 0:16.0, 0:16.0].reshape(3, -1).transpose() | units.m
        dimension = particles.box_counting_dimension()
        self.assertAlmostRelativeEquals(dimension, 3.0, 1)
        # Fractal dimension is scale-free
        particles.position *= 1000
        self.assertAlmostRelativeEquals(dimension, particles.box_counting_dimension(), 10)

        # Particles distributed in the x-y plane
        particles.position = numpy.concatenate((numpy.mgrid[0:64.0, 0:64.0].reshape(2, -1), numpy.zeros((1, 4096)))).transpose() | units.m
        dimension = particles.box_counting_dimension()
        self.assertAlmostRelativeEquals(dimension, 2.0, 1)
        particles.position *= 1000
        self.assertAlmostRelativeEquals(dimension, particles.box_counting_dimension(), 10)

        # Particles distributed along a line
        particles.position = numpy.concatenate([numpy.arange(4096.0).reshape(1, -1)]*3).transpose() | units.m
        dimension = particles.box_counting_dimension()
        self.assertAlmostRelativeEquals(dimension, 1.0, 1)
        particles.position *= 1000
        self.assertAlmostRelativeEquals(dimension, particles.box_counting_dimension(), 10)

        # Particles on a Koch curve
        x, y = self.new_koch_star(level=7)
        numpy.random.seed(123456)
        sel = numpy.random.randint(len(x), size=4096)
        particles.position = numpy.hstack((x[sel], y[sel], numpy.zeros(4096))) | units.m
        dimension = particles.box_counting_dimension()
        self.assertAlmostRelativeEquals(dimension, 1.26186, 1)
        particles.position *= 1000
        self.assertAlmostRelativeEquals(dimension, particles.box_counting_dimension(), 10)

    def test14(self):
        print("Test mass_segregation_from_nearest_neighbour")
        numpy.random.seed(123)
        random.seed(4567)
        number_of_particles = 1000
        particles = new_plummer_sphere(number_of_particles)
        particles.r_squared = particles.position.lengths_squared()
        sorted = particles.sorted_by_attribute("r_squared")

        sorted.mass = numpy.random.uniform(1.0, 2.0, number_of_particles) | nbody_system.mass
        MSR, sigma = sorted.mass_segregation_from_nearest_neighbour(number_of_particles=10, also_compute_uncertainty=True)

        if sys.hexversion > 0x03000000:
            self.assertAlmostEqual(MSR, 1.72632, 3)
            self.assertAlmostEqual(sigma, 0.4127, 3)
        else:
            self.assertAlmostEqual(MSR, 1.7355, 3)
            self.assertAlmostEqual(sigma, 0.3969, 3)

        random.seed(456)
        MSR_of_nonsegregated_systems = []
        for i in range(10):
            sorted.mass = numpy.random.uniform(1.0, 2.0, number_of_particles) | nbody_system.mass
            MSR = sorted.mass_segregation_from_nearest_neighbour(number_of_particles=10, number_of_random_sets=50)
            MSR_of_nonsegregated_systems.append(MSR)
        self.assertAlmostEqual((MSR_of_nonsegregated_systems | units.none).mean(), 1.0, 1)
        self.assertAlmostEqual((MSR_of_nonsegregated_systems | units.none).std(), 0.3, 1)

        sorted.mass = numpy.linspace(2.0, 1.0, number_of_particles) | nbody_system.mass
        MSR, sigma = sorted.mass_segregation_from_nearest_neighbour(number_of_particles=10, number_of_random_sets=20,
            also_compute_uncertainty=True)
        self.assertTrue(MSR > 5.0)

        if sys.hexversion > 0x03000000:
            self.assertAlmostEqual(sigma, 0.3, 1)
        else:
            self.assertAlmostEqual(sigma, 0.4, 1)

    def test15(self):
        scale_R = 1.0 | units.parsec
        scale_M = 1000.0 | units.MSun
        converter = nbody_system.nbody_to_si(scale_M, scale_R)

        for n in range(3162, 3165, 1):
            stars = new_plummer_sphere(
                    n,
                    convert_nbody=converter,
                    )
            stars.mass = (numpy.arange(1, n+1)/(1.*n)) | units.MSun
            potential = stars.potential()

            for i, x in enumerate(stars):
                self.assertAlmostRelativeEqual(potential[i], x.potential())

    def test16(self):
        scale_R = 1.0 | units.parsec
        scale_M = 1000.0 | units.MSun
        converter = nbody_system.nbody_to_si(scale_M, scale_R)

        for n in range(0, 50):
            stars = new_plummer_sphere(
                    50,
                    convert_nbody=converter,
                    )
            stars.mass = numpy.arange(1, 51) | units.MSun
            potential = stars.potential(block_size=n)

            for i, x in enumerate(stars):
                self.assertAlmostRelativeEqual(potential[i], x.potential())


class TestParticlesDomainAttributes(amusetest.TestCase):

    def test1(self):
        particles = Particles(2)
        particles.add_attribute_domain('a')
        particles.add_attribute_domain('b')
        particles.a.foo = 1 | units.m
        particles.b.foo = 2 | units.kg
        particles.foo = 3 | units.s

        self.assertAlmostRelativeEqual(particles.a.foo,  1 | units.m)
        self.assertAlmostRelativeEqual(particles.b.foo,  2 | units.kg)
        self.assertAlmostRelativeEqual(particles.foo,  3 | units.s)

    def test2(self):
        particles = Particles(2)
        particles.add_attribute_domain('a')
        particles.add_attribute_domain('b')
        particles.a.foo = 1 | units.m
        particles.b.bar = 2 | units.kg
        particles.foo = 3 | units.s
        self.assertEqual(
            sorted(particles.a.get_attribute_names_defined_in_store()),
            ['foo']
        )
        self.assertEqual(
            sorted(particles.b.get_attribute_names_defined_in_store()),
            ['bar']
        )

    def test3(self):
        particles = Particles(2)
        particles.add_attribute_domain('a')
        particles.add_attribute_domain('b')
        particles.a.foo = 1 | units.m
        particles.b.foo = 2 | units.kg
        particles.foo = 3 | units.s
        particles.a.add_particle(Particle(foo=2 | units.m))

        self.assertAlmostRelativeEqual(particles.a.foo,  [1, 1, 2] | units.m)
        self.assertAlmostRelativeEqual(particles.b.foo,  [2, 2, 0] | units.kg)
        self.assertAlmostRelativeEqual(particles.foo,  [3, 3, 0] | units.s)

        particles.add_particle(Particle(foo=2 | units.s))
        self.assertAlmostRelativeEqual(particles.a.foo,  [1, 1, 2, 0] | units.m)
        self.assertAlmostRelativeEqual(particles.b.foo,  [2, 2, 0, 0] | units.kg)
        self.assertAlmostRelativeEqual(particles.foo,  [3, 3, 0, 2] | units.s)

    def test4(self):
        particles = Particles(2)
        particles.add_attribute_domain('a')
        particles.add_attribute_domain('b')

        def set_a():
            particles.a = 1 | units.kg
        self.assertRaises(AttributeError, set_a)

    def test5(self):
        particles = Particles(2)
        particles.add_attribute_domain('a')
        particles.add_attribute_domain('b')
        particles.a.foo = 1 | units.m
        particles.b.foo = 2 | units.kg
        particles.foo = 3 | units.s

        particles.a[0].foo = 3 | units.m

        self.assertAlmostRelativeEqual(particles.a.foo,  [3, 1] | units.m)

        particles[0].a.foo = 4 | units.m

        self.assertAlmostRelativeEqual(particles.a.foo,  [4, 1] | units.m)
        self.assertAlmostRelativeEqual(particles.b.foo,  [2, 2] | units.kg)
        self.assertAlmostRelativeEqual(particles.foo,  [3, 3] | units.s)

    def test6(self):
        particles = Particles(2)
        particles.add_attribute_domain('a')
        particles.add_attribute_domain('b')

        def set_a():
            particles[0].a = 1 | units.kg
        self.assertRaises(AttributeError, set_a)



from amuse.units import scaling_converter
from amuse.units import nbody_system

from amuse.test import amusetest


class TestScalingConverter(amusetest.TestCase):

    def test1(self):
        converter = scaling_converter.ScalingConverter(
            length=0.2,
            time=0.1,
        )
        input = 1 | nbody_system.time
        output = converter.convert(input)
        self.assertAlmostRelativeEquals(output, 0.1 | nbody_system.time)

    def test2(self):
        converter = scaling_converter.ScalingConverter(
            length=0.2,
            time=0.1,
        )
        input = 1 | nbody_system.length ** 2
        output = converter.convert(input)
        self.assertAlmostRelativeEquals(output, 0.2 * 0.2 | nbody_system.length ** 2)



import numpy

from amuse.test import amusetest
from amuse.units import nbody_system
from amuse.units import units
from amuse.ic import limepy


class TestLimepy(amusetest.TestCase):
    def test1(self):
        cluster = limepy.Limepy(5, 1, N=100).result

        self.assertAlmostEqual(cluster.total_mass().number, 1.0)
        self.assertEqual(len(cluster), 100)

    def test2(self):
        cluster = limepy.new_limepy_model(7, 2, N=10000)

        self.assertAlmostEqual(cluster.total_mass().number, 1.0)
        self.assertEqual(len(cluster), 10000)

    def test3(self):
        c = nbody_system.nbody_to_si(200 | units.MSun, 2 | units.parsec)
        cluster = limepy.new_limepy_model(7, 2, N=100, converter=c)

        self.assertAlmostEqual(cluster.total_mass(), 200. | units.MSun)
        self.assertEqual(len(cluster), 100)

    def setUp(self):
        if not limepy.scipy_imported:
            self.skip("scipy not installed")



from amuse.test.amusetest import TestWithMPI
from amuse.test import compile_tools
from amuse.test.amusetest import get_amuse_root_dir

# cello
from amuse.community.interface import stopping_conditions
from amuse.support.interface import InCodeComponentImplementation

import os
import shlex

from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.support.exceptions import AmuseException
from amuse.rfi.tools import create_c
from amuse.rfi.tools import create_fortran
from amuse.rfi import channel
from amuse.rfi.core import *
from amuse.community import NO_UNIT
codestring = """
#ifndef NOMPI
#include <mpi.h>
#endif

#include <stopcond.h>
#ifdef __cplusplus
extern "C" {
#endif
int initialize_code()
{
    
    // AMUSE STOPPING CONDITIONS SUPPORT
    supported_conditions = COLLISION_DETECTION_BITMAP | PAIR_DETECTION_BITMAP | TIMEOUT_DETECTION_BITMAP | OUT_OF_BOX_DETECTION_BITMAP;
    // -----------------------
    return 0;
}


int fire_condition(int condition_to_set, int particle_index1, int particle_index2, int rank) {
    int my_rank;
    int error, stopping_index;

#ifndef NOMPI
    error = MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
#else
    error = 0;
    my_rank = rank;
#endif
    if (rank >= 0 && rank != my_rank) { return 0; }
    
    stopping_index = next_index_for_stopping_condition();
    
    error = set_stopping_condition_info(stopping_index, condition_to_set);
    if(particle_index1 > 0) {
        error = set_stopping_condition_particle_index(stopping_index, 0, particle_index1);
    }
    if(particle_index2 >  0) {
        error = set_stopping_condition_particle_index(stopping_index, 1, particle_index2);
    }
    return 0;
}
#ifdef __cplusplus
}
#endif
"""

codestringF = """
      FUNCTION initialize_code()
      IMPLICIT NONE
      include "stopcond.inc"
      INTEGER :: initialize_code
      INTEGER :: set_support_for_condition
      INTEGER :: return
      initialize_code = 0
      return = set_support_for_condition(COLLISION_DETECTION)
      return = set_support_for_condition(PAIR_DETECTION)
      RETURN
      END FUNCTION
"""

codestringFModule = """
MODULE AmuseInterface
CONTAINS
      FUNCTION initialize_code()
          use StoppingConditions
          IMPLICIT NONE
          INTEGER :: initialize_code
          INTEGER :: return
          initialize_code = 0
          return = set_support_for_condition(COLLISION_DETECTION)
          return = set_support_for_condition(PAIR_DETECTION)
      END FUNCTION
      
      FUNCTION fire_condition(condition_to_set, particle_index1, particle_index2, rank)
          use StoppingConditions
          IMPLICIT NONE
          include "mpif.h"
          INTEGER :: fire_condition
          INTEGER :: my_rank
          INTEGER :: error, stopping_index
          INTEGER, intent(in) :: condition_to_set, particle_index1, particle_index2, rank
          fire_condition = 0
          call mpi_comm_rank(MPI_COMM_WORLD, my_rank, error)
          
          if (rank.GE.0 .AND. rank.NE.my_rank) then
            return
          end if
          stopping_index = next_index_for_stopping_condition()
          error = set_stopping_condition_info(stopping_index, condition_to_set)
          if(particle_index1 .GT.  0) then
            error = set_stopping_condition_particle_index(stopping_index, 0, particle_index1)
          end if
          if(particle_index2 .GT.  0) then
            error = set_stopping_condition_particle_index(stopping_index, 1, particle_index2)
          end if
      END FUNCTION
END MODULE
"""


class ForTestingInterface(CodeInterface, stopping_conditions.StoppingConditionInterface):
    def __init__(self, exefile, **options):
        CodeInterface.__init__(self, exefile, **options)

    include_headers = ['stopcond.h']

    @legacy_function
    def initialize_code():
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        function.can_handle_array = False
        return function

    @legacy_function
    def reset_stopping_conditions():
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        function.can_handle_array = False
        return function

    @legacy_function
    def next_index_for_stopping_condition():
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        function.result_unit = NO_UNIT
        function.can_handle_array = False
        return function

    @legacy_function
    def set_stopping_condition_info():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('index_of_the_condition', dtype='int32', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def set_stopping_condition_particle_index():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('index_of_the_condition', dtype='int32', direction=function.IN)
        function.addParameter('index_of_the_particle', dtype='int32', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def mpi_setup_stopping_conditions():
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        function.can_handle_array = False
        return function

    @legacy_function
    def mpi_collect_stopping_conditions():
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        function.can_handle_array = False
        return function

    @legacy_function
    def mpi_distribute_stopping_conditions():
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        function.can_handle_array = False
        return function


class ForTestingInterfaceFortranModule(ForTestingInterface):
    use_modules = ['StoppingConditions', 'AmuseInterface']

    @legacy_function
    def fire_condition():
        function = LegacyFunctionSpecification()
        function.addParameter('condition_to_set', dtype='int32', direction=function.IN)
        function.addParameter('particle_index_1', dtype='int32', direction=function.IN)
        function.addParameter('particle_index_2', dtype='int32', direction=function.IN)
        function.addParameter('rank', dtype='int32', direction=function.IN, default=-1)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function


class ForTesting(InCodeComponentImplementation):
    def __init__(self, exefile, **options):
        if 'community_interface' in options:
            interface = options['community_interface']
        else:
            interface = ForTestingInterface
        self.stopping_conditions = stopping_conditions.StoppingConditions(self)
        InCodeComponentImplementation.__init__(self, interface(exefile, **options), **options)
        self.my_particles = datamodel.Particles()

    def define_methods(self, object):
        self.stopping_conditions.define_methods(object)

    def new_particle(self, mass):
        particles = datamodel.Particles(len(mass))
        particles.mass = mass
        self.my_particles.add_particles(particles)
        return list(range(len(self.my_particles)-len(mass), len(self.my_particles)))

    def get_mass(self, indices):
        return self.my_particles.mass[indices]

    def delete_particle(self, particle):
        self.my_particles.remove_particle(particle)

    def define_particle_sets(self, object):
        object.define_set('particles', 'index_of_the_particle')
        object.set_new('particles', 'new_particle')
        object.set_delete('particles', 'delete_particle')
        object.add_getter('particles', 'get_mass', names=("mass",))
        self.stopping_conditions.define_particle_set(object)


class _AbstractTestInterface(TestWithMPI):

    @classmethod
    def get_libname(cls):
        return "stopcond"

    @classmethod
    def setup_class(cls):
        cls.check_can_compile_modules()
        cls.exefile = compile_tools.build_worker(codestring,
            cls.get_path_to_results(),
            cls.get_interface_class(), write_header=False,
            extra_args=["-L"+get_amuse_root_dir()+"/lib/stopcond", "-l" + cls.get_libname()]
            )

    @classmethod
    def get_interface_class(cls):
        return ForTestingInterface


class TestInterface(_AbstractTestInterface):

    def test1(self):
        # ~ print self.exefile
        instance = ForTestingInterface(self.exefile)
        instance.reset_stopping_conditions()
        next = instance.next_index_for_stopping_condition()
        next = instance.next_index_for_stopping_condition()
        instance.stop()
        self.assertEqual(next, 1)

    def test2(self):
        instance = ForTesting(self.exefile)  # , debugger = "xterm")
        instance.initialize_code()
        self.assertTrue(instance.stopping_conditions.pair_detection.is_supported())
        self.assertTrue(instance.stopping_conditions.collision_detection.is_supported())
        self.assertFalse(instance.stopping_conditions.escaper_detection.is_supported())
        instance.stop()

    def test3(self):
        instance = ForTesting(self.exefile)  # , debugger = "xterm")
        instance.initialize_code()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_enabled())
        instance.stopping_conditions.pair_detection.enable()
        self.assertTrue(instance.stopping_conditions.pair_detection.is_enabled())
        instance.stopping_conditions.pair_detection.disable()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_enabled())
        instance.stop()

    def test4(self):
        instance = ForTesting(self.exefile)
        instance.reset_stopping_conditions()
        next = instance.next_index_for_stopping_condition()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_set())
        # ~ print next,instance.stopping_conditions.pair_detection.type
        instance.set_stopping_condition_info(next, instance.stopping_conditions.pair_detection.type)

        self.assertTrue(instance.stopping_conditions.pair_detection.is_set())
        instance.stop()

    def test5(self):
        instance = ForTesting(self.exefile)
        instance.reset_stopping_conditions()
        next = instance.next_index_for_stopping_condition()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_set())

        instance.set_stopping_condition_info(next, instance.stopping_conditions.pair_detection.type)
        instance.set_stopping_condition_particle_index(next, 0, 11)
        instance.set_stopping_condition_particle_index(next, 1, 12)
        self.assertTrue(instance.stopping_conditions.pair_detection.is_set())
        self.assertEqual(11, instance.get_stopping_condition_particle_index(next, 0))
        self.assertEqual(12, instance.get_stopping_condition_particle_index(next, 1))
        instance.stop()

    def test6(self):
        instance = ForTesting(self.exefile)
        instance.reset_stopping_conditions()
        next = instance.next_index_for_stopping_condition()
        instance.set_stopping_condition_info(next, instance.stopping_conditions.out_of_box_detection.type)
        self.assertTrue(instance.stopping_conditions.out_of_box_detection.is_set())
        instance.stop()

    def test7(self):
        instance = ForTesting(self.exefile)
        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)
        instance.reset_stopping_conditions()

        pairs = [(11, 12), (0, 4), (3, 18), (7, 2)]
        next = instance.next_index_for_stopping_condition()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_set())
        instance.set_stopping_condition_info(next, instance.stopping_conditions.pair_detection.type)
        instance.set_stopping_condition_particle_index(next, 0, pairs[0][0])
        instance.set_stopping_condition_particle_index(next, 1, pairs[0][1])
        self.assertEqual(11, instance.get_stopping_condition_particle_index(next, 0))
        self.assertEqual(12, instance.get_stopping_condition_particle_index(next, 1))
        self.assertTrue(instance.stopping_conditions.pair_detection.is_set())
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(0)), 1)
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(1)), 1)

        for index1, index2 in pairs[1:]:
            next = instance.next_index_for_stopping_condition()
            instance.set_stopping_condition_info(next, instance.stopping_conditions.pair_detection.type)
            instance.set_stopping_condition_particle_index(next, 0, index1)
            instance.set_stopping_condition_particle_index(next, 1, index2)
            self.assertEqual(index1, instance.get_stopping_condition_particle_index(next, 0))
            self.assertEqual(index2, instance.get_stopping_condition_particle_index(next, 1))
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(0)), 4)
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(1)), 4)
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(2)), 0)

        self.assertEqual(instance.stopping_conditions.pair_detection.particles(0).mass,
            [first + 1 for first, second in pairs] | units.kg)
        self.assertEqual(instance.stopping_conditions.pair_detection.particles(1).mass,
            [second + 1 for first, second in pairs] | units.kg)
        instance.stop()

    def test8(self):
        instance = ForTesting(self.exefile)
        instance.initialize_code()
        self.assertFalse(instance.stopping_conditions.escaper_detection.is_supported())
        self.assertRaises(AmuseException, instance.stopping_conditions.escaper_detection.enable, expected_message="Can't enable stopping condition 'escaper_detection', since 'ForTesting' does not support this condition.")
        instance.stop()

    def test9(self):
        instance = ForTestingInterface(self.exefile)
        instance.reset_stopping_conditions()
        nmax = 2048
        for i in range(nmax):
            next = instance.next_index_for_stopping_condition()
            # ~ print i, next
            self.assertEqual(next, i)
        instance.stop()


class TestInterfaceMP(_AbstractTestInterface):

    @classmethod
    def get_interface_class(self):
        return ForTestingInterfaceFortranModule

    def get_number_of_workers(self):
        return 3

    @classmethod
    def get_libname(self):
        return "stopcondmpi"

    def test1(self):
        number_of_workers = 4
        instance = ForTestingInterface(self.exefile, number_of_workers=number_of_workers)
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()
        instance.enable_stopping_condition(1)
        nmax = 50
        for i in range(nmax):
            next = instance.next_index_for_stopping_condition()
            self.assertEqual(next, i)
        i, error = instance.get_number_of_stopping_conditions_set()
        self.assertEqual(error, 0)
        self.assertEqual(i, nmax)
        instance.mpi_collect_stopping_conditions()
        i, error = instance.get_number_of_stopping_conditions_set()
        self.assertEqual(error, 0)
        self.assertEqual(i, number_of_workers * nmax)

        instance.stop()

    def test2(self):
        instance = ForTesting(
            self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=self.get_number_of_workers()
        )
        instance.initialize_code()
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()

        pair_detection = instance.stopping_conditions.pair_detection

        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)

        instance.stopping_conditions.pair_detection.enable()

        instance.mpi_distribute_stopping_conditions()

        # ~ print pair_detection.type
        instance.fire_condition(
            pair_detection.type,
            1, 2, -1
        )
        instance.mpi_collect_stopping_conditions()
        self.assertTrue(pair_detection.is_set())
        self.assertEqual(len(pair_detection.particles(0)), self.get_number_of_workers())
        self.assertEqual(len(pair_detection.particles(1)), self.get_number_of_workers())
        self.assertEqual(pair_detection.particles(0).key, particles[1].key)
        self.assertEqual(pair_detection.particles(1).key, particles[2].key)
        self.assertEqual(pair_detection.particles(0).mass, [2, 2, 2] | units.kg)
        self.assertEqual(pair_detection.particles(1).mass, [3, 3, 3] | units.kg)
        instance.stop()

    def test5(self):
        instance = ForTesting(
            self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=self.get_number_of_workers()
        )
        instance.initialize_code()
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()

        pair_detection = instance.stopping_conditions.pair_detection

        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)

        instance.stopping_conditions.pair_detection.enable()

        instance.mpi_distribute_stopping_conditions()
        for rank in range(self.get_number_of_workers()):
            # ~ print pair_detection.type
            instance.fire_condition(
                pair_detection.type,
                1, 2, rank
            )
            instance.mpi_collect_stopping_conditions()
            self.assertTrue(pair_detection.is_set())
            self.assertEqual(len(pair_detection.particles(0)), 1)
            self.assertEqual(len(pair_detection.particles(1)), 1)
            self.assertEqual(pair_detection.particles(0).key, particles[1].key)
            self.assertEqual(pair_detection.particles(1).key, particles[2].key)
            self.assertEqual(pair_detection.particles(0).mass, [2] | units.kg)
            self.assertEqual(pair_detection.particles(1).mass, [3] | units.kg)
            instance.reset_stopping_conditions()
            instance.stopping_conditions.pair_detection.enable()

        instance.stop()

    def test3(self):
        instance = ForTesting(
            self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=self.get_number_of_workers()
        )
        instance.initialize_code()
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()

        pair_detection = instance.stopping_conditions.pair_detection

        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)

        instance.stopping_conditions.pair_detection.enable()

        instance.mpi_distribute_stopping_conditions()

        instance.fire_condition(
            pair_detection.type,
            1, 2, 0
        )
        instance.fire_condition(
            pair_detection.type,
            3, 4, 1
        )
        instance.fire_condition(
            pair_detection.type,
            5, 6, 2
        )
        instance.mpi_collect_stopping_conditions()
        self.assertTrue(pair_detection.is_set())
        self.assertEqual(len(pair_detection.particles(0)), 3)
        self.assertEqual(len(pair_detection.particles(1)), 3)
        self.assertEqual(pair_detection.particles(0).key[0], particles[1].key)
        self.assertEqual(pair_detection.particles(1).key[0], particles[2].key)
        self.assertEqual(pair_detection.particles(0).key[1], particles[3].key)
        self.assertEqual(pair_detection.particles(1).key[1], particles[4].key)
        self.assertEqual(pair_detection.particles(0).key[2], particles[5].key)
        self.assertEqual(pair_detection.particles(1).key[2], particles[6].key)
        instance.reset_stopping_conditions()
        instance.stopping_conditions.pair_detection.enable()

        instance.stop()

    def test4(self):
        instance = ForTesting(
            self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=self.get_number_of_workers()
        )
        instance.initialize_code()
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()

        pair_detection = instance.stopping_conditions.pair_detection

        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)

        instance.stopping_conditions.pair_detection.enable()

        instance.mpi_collect_stopping_conditions()

        instance.fire_condition(
            pair_detection.type,
            -1, -1, -1
        )
        instance.mpi_distribute_stopping_conditions()
        self.assertTrue(pair_detection.is_set())
        self.assertEqual(len(pair_detection.particles(0)), 0)

        instance.stop()


class _AbstractTestInterfaceFortran:

    @classmethod
    def get_libname(cls):
        return 'stopcond'

    @classmethod
    def get_mpidir(cls):
        return ''

    @classmethod
    def get_codestring(cls):
        return codestringF

    @classmethod
    def get_interface_class(cls):
        return ForTestingInterface

    def get_number_of_workers(self):
        return 1

    @classmethod
    def setup_class(cls):
        cls.check_can_compile_modules()
        cls.exefile = compile_tools.build_fortran_worker(cls.get_codestring(),
            cls.get_path_to_results(), cls.get_interface_class(), needs_mpi=True,
            extra_fflags=["-I", "{0}/lib/stopcond".format(get_amuse_root_dir())],
            extra_ldflags=["-L{0}/lib/stopcond".format(get_amuse_root_dir()), "-l"+cls.get_libname()])


class _TestInterfaceFortranSingleProcess(TestWithMPI, _AbstractTestInterfaceFortran):

    def get_number_of_workers(self):
        return 1

    def test1(self):
        instance = ForTestingInterface(self.exefile, number_of_workers=self.get_number_of_workers())
        instance.reset_stopping_conditions()
        next = instance.next_index_for_stopping_condition()
        next = instance.next_index_for_stopping_condition()
        instance.stop()
        self.assertEqual(next, 1)

    def test2(self):
        instance = ForTesting(self.exefile, number_of_workers=self.get_number_of_workers())  # , debugger = "xterm")
        instance.initialize_code()
        self.assertTrue(instance.stopping_conditions.pair_detection.is_supported())
        self.assertTrue(instance.stopping_conditions.collision_detection.is_supported())
        self.assertFalse(instance.stopping_conditions.escaper_detection.is_supported())
        instance.stop()

    def test3(self):
        instance = ForTesting(self.exefile, number_of_workers=self.get_number_of_workers())  # , debugger = "xterm")
        instance.initialize_code()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_enabled())
        instance.stopping_conditions.pair_detection.enable()
        self.assertTrue(instance.stopping_conditions.pair_detection.is_enabled())
        instance.stopping_conditions.pair_detection.disable()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_enabled())
        instance.stop()

    def test4(self):
        instance = ForTesting(self.exefile, number_of_workers=self.get_number_of_workers())
        instance.reset_stopping_conditions()

        next = instance.next_index_for_stopping_condition()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_set())

        instance.set_stopping_condition_info(next, instance.stopping_conditions.pair_detection.type)

        self.assertTrue(instance.stopping_conditions.pair_detection.is_set())
        instance.stop()

    def test5(self):
        instance = ForTesting(self.exefile, number_of_workers=self.get_number_of_workers())
        instance.reset_stopping_conditions()
        next = instance.next_index_for_stopping_condition()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_set())

        instance.set_stopping_condition_info(next, instance.stopping_conditions.pair_detection.type)
        instance.set_stopping_condition_particle_index(next, 0, 11)
        instance.set_stopping_condition_particle_index(next, 1, 12)
        self.assertTrue(instance.stopping_conditions.pair_detection.is_set())
        self.assertEqual(11, instance.get_stopping_condition_particle_index(next, 0))
        self.assertEqual(12, instance.get_stopping_condition_particle_index(next, 1))
        instance.stop()

    def test6(self):
        instance = ForTesting(self.exefile, number_of_workers=self.get_number_of_workers())
        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)
        instance.reset_stopping_conditions()

        pairs = [(11, 12), (0, 4), (3, 18), (7, 2)]
        next = instance.next_index_for_stopping_condition()
        self.assertFalse(instance.stopping_conditions.pair_detection.is_set())
        instance.set_stopping_condition_info(next, instance.stopping_conditions.pair_detection.type)
        instance.set_stopping_condition_particle_index(next, 0, pairs[0][0])
        instance.set_stopping_condition_particle_index(next, 1, pairs[0][1])
        self.assertEqual(11, instance.get_stopping_condition_particle_index(next, 0))
        self.assertEqual(12, instance.get_stopping_condition_particle_index(next, 1))
        self.assertTrue(instance.stopping_conditions.pair_detection.is_set())
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(0)), 1)
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(1)), 1)

        for index1, index2 in pairs[1:]:
            next = instance.next_index_for_stopping_condition()
            instance.set_stopping_condition_info(next, instance.stopping_conditions.pair_detection.type)
            instance.set_stopping_condition_particle_index(next, 0, index1)
            instance.set_stopping_condition_particle_index(next, 1, index2)
            self.assertEqual(index1, instance.get_stopping_condition_particle_index(next, 0))
            self.assertEqual(index2, instance.get_stopping_condition_particle_index(next, 1))
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(0)), 4)
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(1)), 4)
        self.assertEqual(len(instance.stopping_conditions.pair_detection.particles(2)), 0)

        self.assertEqual(instance.stopping_conditions.pair_detection.particles(0).mass,
            [first + 1 for first, second in pairs] | units.kg)
        self.assertEqual(instance.stopping_conditions.pair_detection.particles(1).mass,
            [second + 1 for first, second in pairs] | units.kg)
        instance.stop()

    def test8(self):
        instance = ForTesting(self.exefile, number_of_workers=self.get_number_of_workers())
        instance.initialize_code()
        self.assertFalse(instance.stopping_conditions.escaper_detection.is_supported())
        self.assertRaises(AmuseException, instance.stopping_conditions.escaper_detection.enable, expected_message="Can't enable stopping condition 'escaper_detection', since 'ForTesting' does not support this condition.")
        instance.stop()

    def test9(self):
        instance = ForTestingInterface(self.exefile, number_of_workers=self.get_number_of_workers())
        instance.initialize_code()
        instance.reset_stopping_conditions()
        nmax = 2048
        for i in range(nmax):
            next = instance.next_index_for_stopping_condition()
            # ~ print i, next
            self.assertEqual(next, i)
        instance.stop()


class TestInterfaceFortran(_TestInterfaceFortranSingleProcess):

    @classmethod
    def get_libname(cls):
        return 'stopcond'

    @classmethod
    def get_codestring(cls):
        return codestringF

    @classmethod
    def get_interface_class(cls):
        return ForTestingInterface


class TestInterfaceFortranModule(_TestInterfaceFortranSingleProcess):

    @classmethod
    def get_libname(cls):
        return 'stopcond'

    @classmethod
    def get_codestring(cls):
        return codestringFModule

    @classmethod
    def get_interface_class(cls):
        return ForTestingInterfaceFortranModule


class TestInterfaceFortranModuleMultiprocess(TestWithMPI, _AbstractTestInterfaceFortran):

    @classmethod
    def get_libname(cls):
        return 'stopcondmpi'

    @classmethod
    def get_codestring(cls):
        return codestringFModule

    @classmethod
    def get_interface_class(cls):
        return ForTestingInterfaceFortranModule

    def get_number_of_workers(self):
        return 3

    @classmethod
    def get_mpidir(self):
        return ''

    def test1(self):
        instance = ForTesting(
            self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=self.get_number_of_workers()
        )
        instance.initialize_code()
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()

        pair_detection = instance.stopping_conditions.pair_detection

        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)

        instance.stopping_conditions.pair_detection.enable()

        instance.mpi_distribute_stopping_conditions()

        # ~ print pair_detection.type
        instance.fire_condition(
            pair_detection.type,
            1, 2, -1
        )
        instance.mpi_collect_stopping_conditions()
        self.assertTrue(pair_detection.is_set())
        self.assertEqual(len(pair_detection.particles(0)), self.get_number_of_workers())
        self.assertEqual(len(pair_detection.particles(1)), self.get_number_of_workers())
        self.assertEqual(pair_detection.particles(0).key, particles[1].key)
        self.assertEqual(pair_detection.particles(1).key, particles[2].key)
        self.assertEqual(pair_detection.particles(0).mass, [2, 2, 2] | units.kg)
        self.assertEqual(pair_detection.particles(1).mass, [3, 3, 3] | units.kg)
        instance.stop()

    def test2(self):
        instance = ForTesting(
            self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=self.get_number_of_workers()
        )
        instance.initialize_code()
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()

        pair_detection = instance.stopping_conditions.pair_detection

        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)

        instance.stopping_conditions.pair_detection.enable()

        instance.mpi_distribute_stopping_conditions()
        for rank in range(self.get_number_of_workers()):
            # ~ print pair_detection.type
            instance.fire_condition(
                pair_detection.type,
                1, 2, rank
            )
            instance.mpi_collect_stopping_conditions()
            self.assertTrue(pair_detection.is_set())
            self.assertEqual(len(pair_detection.particles(0)), 1)
            self.assertEqual(len(pair_detection.particles(1)), 1)
            self.assertEqual(pair_detection.particles(0).key, particles[1].key)
            self.assertEqual(pair_detection.particles(1).key, particles[2].key)
            self.assertEqual(pair_detection.particles(0).mass, [2] | units.kg)
            self.assertEqual(pair_detection.particles(1).mass, [3] | units.kg)
            instance.reset_stopping_conditions()
            instance.stopping_conditions.pair_detection.enable()

        instance.stop()

    def test3(self):
        instance = ForTesting(
            self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=self.get_number_of_workers()
        )
        instance.initialize_code()
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()

        pair_detection = instance.stopping_conditions.pair_detection

        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)

        instance.stopping_conditions.pair_detection.enable()

        instance.mpi_distribute_stopping_conditions()

        instance.fire_condition(
            pair_detection.type,
            1, 2, 0
        )
        instance.fire_condition(
            pair_detection.type,
            3, 4, 1
        )
        instance.fire_condition(
            pair_detection.type,
            5, 6, 2
        )
        instance.mpi_collect_stopping_conditions()
        self.assertTrue(pair_detection.is_set())
        self.assertEqual(len(pair_detection.particles(0)), 3)
        self.assertEqual(len(pair_detection.particles(1)), 3)
        self.assertEqual(pair_detection.particles(0).key[0], particles[1].key)
        self.assertEqual(pair_detection.particles(1).key[0], particles[2].key)
        self.assertEqual(pair_detection.particles(0).key[1], particles[3].key)
        self.assertEqual(pair_detection.particles(1).key[1], particles[4].key)
        self.assertEqual(pair_detection.particles(0).key[2], particles[5].key)
        self.assertEqual(pair_detection.particles(1).key[2], particles[6].key)
        instance.reset_stopping_conditions()
        instance.stopping_conditions.pair_detection.enable()

        instance.stop()

    def test4(self):
        instance = ForTesting(
            self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=self.get_number_of_workers()
        )
        instance.initialize_code()
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()

        pair_detection = instance.stopping_conditions.pair_detection

        particles = datamodel.Particles(20)
        particles.mass = range(1, 21) | units.kg
        instance.particles.add_particles(particles)

        instance.stopping_conditions.pair_detection.enable()

        instance.mpi_collect_stopping_conditions()

        instance.fire_condition(
            pair_detection.type,
            -1, -1, -1
        )
        instance.mpi_distribute_stopping_conditions()
        self.assertTrue(pair_detection.is_set())
        self.assertEqual(len(pair_detection.particles(0)), 0)

        instance.stop()

    def test5(self):
        number_of_workers = 4
        instance = ForTestingInterface(self.exefile,
            community_interface=ForTestingInterfaceFortranModule,
            number_of_workers=number_of_workers)
        instance.reset_stopping_conditions()
        instance.mpi_setup_stopping_conditions()
        instance.enable_stopping_condition(1)
        nmax = 50
        for i in range(nmax):
            next = instance.next_index_for_stopping_condition()
            self.assertEqual(next, i)
        i, error = instance.get_number_of_stopping_conditions_set()
        self.assertEqual(error, 0)
        self.assertEqual(i, nmax)
        instance.mpi_collect_stopping_conditions()
        i, error = instance.get_number_of_stopping_conditions_set()
        self.assertEqual(error, 0)
        self.assertEqual(i, number_of_workers * nmax)

        instance.stop()



from amuse.test import amusetest

import pickle

from amuse.support.exceptions import AmuseException

from amuse.units import core
from amuse.units import si
from amuse.units import nbody_system
from amuse.units import generic_unit_system
from amuse.units.quantities import zero
from amuse.units.units import *
from amuse.units.constants import *

from amuse.datamodel import Particles, parameters

import subprocess
import pickle
import sys
import os


class TestPicklingOfUnitsAndQuantities(amusetest.TestCase):

    def test1(self):
        km = 1000 * m
        self.assertEqual(1000, km.value_in(m))
        pickled_km = pickle.dumps(km)
        unpickled_km = pickle.loads(pickled_km)
        self.assertEqual(1000, unpickled_km.value_in(m))

    def test2(self):
        km = 1000 * m
        quantity = 12.0 | km
        self.assertEqual(12000, quantity.value_in(m))
        pickled_quantity = pickle.dumps(quantity)
        unpickled_quantity = pickle.loads(pickled_quantity)
        self.assertEqual(12000, unpickled_quantity.value_in(m))
        self.assertEqual(quantity, unpickled_quantity)

    def test3(self):
        pickled_si_sytem = pickle.dumps(si.system)
        unpickled_si_sytem = pickle.loads(pickled_si_sytem)
        self.assertTrue(unpickled_si_sytem is si.system)

    def test4(self):
        quantity = 12.0 | nbody_system.energy
        pickled_quantity = pickle.dumps(quantity)
        unpickled_quantity = pickle.loads(pickled_quantity)
        self.assertEqual(quantity, unpickled_quantity)

    def test5(self):
        quantity = 12.0 | parsec
        pickled_quantity = pickle.dumps(quantity)
        unpickled_quantity = pickle.loads(pickled_quantity)
        self.assertEqual(quantity, unpickled_quantity)
        self.assertEqual(str(quantity), str(unpickled_quantity))
        self.assertEqual(12.0, unpickled_quantity.value_in(parsec))

    def test6(self):
        quantity = [12.0, 15.0] | parsec
        pickled_quantity = pickle.dumps(quantity)
        unpickled_quantity = pickle.loads(pickled_quantity)
        self.assertEqual(quantity, unpickled_quantity)
        self.assertEqual(str(quantity), str(unpickled_quantity))

    def test7(self):
        quantity = zero
        pickled_quantity = pickle.dumps(quantity)
        unpickled_quantity = pickle.loads(pickled_quantity)
        self.assertEqual(quantity, unpickled_quantity)
        self.assertTrue(quantity is unpickled_quantity)
        self.assertEqual(str(quantity), str(unpickled_quantity))

    def test8(self):
        quantity = 1 | nbody_system.time
        pickled_quantity = pickle.dumps(quantity)
        unpickled_quantity = pickle.loads(pickled_quantity)
        self.assertEqual(quantity, unpickled_quantity)
        self.assertEqual(str(quantity), str(unpickled_quantity))

    def test9(self):
        quantity = 1.3 | nbody_system.time
        path = os.path.abspath(os.path.join(self.get_path_to_results(), "test9.pickle"))

        with open(path, "wb") as stream:
            pickle.dump(quantity, stream)

        pythonpath = os.pathsep.join(sys.path)
        env = os.environ.copy()
        env['PYTHONPATH'] = pythonpath
        code = "import pickle;stream = open('{0}', 'rb'); print(str(pickle.load(stream)));stream.close()".format(path)

        process = subprocess.Popen([
                sys.executable,
                "-c",
                code,
               "--no-report-references"
            ], stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
        )
        unpickled_quantity_string, error_string = process.communicate()
        self.assertEqual(process.returncode, 0)
        self.assertEqual(str(quantity), unpickled_quantity_string.strip().decode('utf-8'))

    def test10(self):
        quantity = 1 | parsec
        path = os.path.abspath(os.path.join(self.get_path_to_results(), "test10.pickle"))
        with open(path, "wb") as stream:
            pickle.dump(quantity, stream)

        pythonpath = os.pathsep.join(sys.path)
        env = os.environ.copy()
        env['PYTHONPATH'] = pythonpath
        code = "import pickle;stream = open('{0}', 'rb'); print(str(pickle.load(stream)));stream.close()".format(path)

        process = subprocess.Popen([
                sys.executable,
                "-c",
               code,
               "--no-report-references"
            ], stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
        )
        unpickled_quantity_string, error_string = process.communicate()
        self.assertEqual(process.returncode, 0)
        self.assertEqual(str(quantity), unpickled_quantity_string.strip().decode('utf-8'))

    def test11(self):
        value = 1 | stellar_type
        self.assertEqual(1 | stellar_type, value)
        pickled = pickle.dumps(value)
        unpickled_value = pickle.loads(pickled)
        self.assertEqual(1 | stellar_type, unpickled_value)


class TestPicklingOfParticleSets(amusetest.TestCase):

    def test1(self):
        particles = Particles(4)
        particles.mass = [1, 2, 3, 4] | km
        pickled_particles = pickle.dumps(particles)
        unpickled_particles = pickle.loads(pickled_particles)
        self.assertAlmostRelativeEquals(unpickled_particles.mass, [1, 2, 3, 4] | km)

    def test2(self):
        particles = Particles(4)
        particles.mass = [1, 2, 3, 6] | kg
        particles.position = [[0, 0, 0], [3, 0, 0], [0, 4, 0], [3, 4, 0]] | m
        self.assertEqual(particles.center_of_mass(), [2, 3, 0] | m)
        pickled_particles = pickle.dumps(particles)
        unpickled_particles = pickle.loads(pickled_particles)
        self.assertAlmostRelativeEquals(unpickled_particles.mass, [1, 2, 3, 6] | kg)
        self.assertEqual(unpickled_particles.center_of_mass(), [2, 3, 0] | m)

    def test3(self):
        particles = Particles(4)
        particles.mass = [1, 2, 3, 6] | kg
        particles.position = [[0, 0, 0], [3, 0, 0], [0, 4, 0], [3, 4, 0]] | m
        self.assertEqual(particles.center_of_mass(), [2, 3, 0] | m)
        pickled_particles = pickle.dumps(particles)
        unpickled_particles = pickle.loads(pickled_particles)
        pickled_particles = pickle.dumps(particles)  # dump it twice!
        unpickled_particles = pickle.loads(pickled_particles)
        self.assertAlmostRelativeEquals(unpickled_particles.mass, [1, 2, 3, 6] | kg)
        self.assertEqual(unpickled_particles.center_of_mass(), [2, 3, 0] | m)

    def test4(self):
        particles = Particles(4)
        particles.mass = [1, 2, 3, 6] | kg
        particles.position = [[0, 0, 0], [3, 0, 0], [0, 4, 0], [3, 4, 0]] | m
        pickled_particles = pickle.dumps(list(particles))
        unpickled_particles = pickle.loads(pickled_particles)
        self.assertEqual(len(unpickled_particles), 4)
        unpickled_particles = Particles(particles=unpickled_particles)
        self.assertAlmostRelativeEquals(unpickled_particles.mass, [1, 2, 3, 6] | kg)
        self.assertEqual(unpickled_particles.center_of_mass(), [2, 3, 0] | m)


class BaseTestModule(object):
    def before_get_parameter(self):
        return

    def before_set_parameter(self):
        return


class TestPicklingOfParameters(amusetest.TestCase):

    def test1(self):
        definition = parameters.ModuleMethodParameterDefinition(
            "get_test",
            "set_test",
            "test_name",
            "a test parameter",
            0.1 | m
        )

        class TestModule(BaseTestModule):
            def get_test(self):
                return self.x

            def set_test(self, value):
                self.x = value

        o = TestModule()
        set = parameters.Parameters([definition,], o)
        set.test_name = 10 | m

        self.assertEqual(o.x, 10 | m)
        self.assertEqual(set.test_name, 10 | m)

        memento = set.copy()
        self.assertEqual(memento.test_name, 10 | m)

        pickled_memento = pickle.dumps(memento)
        unpickled_memento = pickle.loads(pickled_memento)

        self.assertEqual(memento.test_name, unpickled_memento.test_name)
